#!/usr/bin/env python
"""
JSON schema to relatively pretty HTML
"""
import argparse
import json
import logging
import os
import re
import six
import sys

_h = logging.StreamHandler()
_f = logging.Formatter(fmt='%(asctime)s %(name)s [%(levelname)s] %(message)s')
_h.setFormatter(_f)
_log = logging.getLogger(__name__)
_log.addHandler(_h)


DEFAULT_CSS = '''
'''


def get_file(file_or_path, mode='r'):
    """Open a file for reading, or simply return the file object.
    """
    if hasattr(file_or_path, 'read'):
        return file_or_path
    return open(file_or_path, mode)


class SchemaProcessor(object):
    def __init__(self, schema_file, schema_events):
        schema = json.load(get_file(schema_file))
        assert schema['type'] == 'object', 'Top-level "type" != "object"'
        self.defs = schema.get('definitions', {})
        self.events = schema_events
        self._schema = schema

    def process(self):
        status = 0
        try:
            self._parse_object(self._schema, '')
        except Exception as err:
            _log.fatal('While processing schema: {}'.format(err))
            status = -1
        self.events.done()
        return status

    def _dereference(self, ref):
        # print("@@ dereference: {}".format(ref))
        m = re.match('#/definitions/(.*)', ref)
        if m is None:
            raise ValueError('Unrecognized reference:'
                             'expected format "#/definitions/<name>", got: '
                             '"{}"'.format(ref))
        defname = m.group(1)
        if defname not in self.defs:
            raise KeyError('Cannot find definition "{}" from reference "{}"'
                           .format(defname, ref))
        deref = self.defs[defname]
        # print('@@ END dereference {} = {}'.format(ref, deref))
        return deref

    def _parse_object(self, obj, name):
        # print('@@ parse object {}: {}'.format(name, obj))
        if '$ref' in obj:
            self._parse_object(self._dereference(obj['$ref']), name)
        else:
            self.events.begin_object(name)
            if 'description' in obj:
                self.events.description(obj['description'])
            props = obj.get('properties', {})
            self.events.begin_children()
            for k, v in six.iteritems(props):
                if k not in ('description', 'type'):
                    self.events.begin_child()
                    self._parse_child(k, v)
                    self.events.end_child()
            self.events.end_children()
            self.events.end_object()

    def _parse_child(self, name, val):
        if '$ref' in val:
            val = self._dereference(val['$ref'])
        otype = val['type']
        self._parse_type(name, otype, val)

    def _parse_type(self, name, otype, val):
        if otype == 'object':
            self._parse_object(val, name)
        elif otype == 'array':
            self._parse_array(val, name)
        else:
            self._parse_scalar(name, otype, val)

    def _parse_array(self, arr, name):
        # print("@@ parse array: {}".format(name))
        self.events.begin_array(name)
        self.events.description(arr.get('description', ''))
        items = arr['items']
        if '$ref' in items:
            items = self._dereference(items['$ref'])
        self.events.begin_array_items()
        if isinstance(items, list):
            for i, item in enumerate(items):
                self._parse_type(str(i), item['type'], item)
        else:
            self._parse_type(None, items['type'], items)
        self.events.end_array_items()
        self.events.end_array()
        # print("@@ END parse array: {}".format(name))

    def _parse_scalar(self, name, typename, val):
        # print("@@ parse scalar: {}".format(name))
        self.events.begin_node(name)
        self.events.typename(typename)
        if 'description' in val:
            self.events.description(val['description'])
        if 'examples' in val:
            self.events.examples(val['examples'])
        elif 'enum' in val:
            self.events.enum(val['enum'])
        self.events.end_node()
        # print("@@ END parse scalar: {}".format(name))


class HtmlSchemaEvents(object):
    def __init__(self, output_file, css_input_file, title='',
                 embeddable=False):
        self.f = get_file(output_file, mode='w')
        if css_input_file:
            self.css = get_file(css_input_file).read()
        else:
            self.css = DEFAULT_CSS
        self._indent = 0
        self.title = title
        self._embed = embeddable
        self.begin()

    def begin(self):
        if not self._embed:
            self.f.write('<!DOCTYPE html>\n'
                         '<html lang="en">\n')
            self.f.write('<head>\n')
            # Grab Google font(s) [note: as of 8/2017 this worked]
            self.f.write('<link rel="stylesheet" type="text/css" '
                         'href="http://fonts.googleapis.com/css'
                         '?family=Open+Sans">')
            # add inline CSS
            self.f.write('<style>\n{}\n</style>'.format(self.css))
            # add page metadata
            self.f.write('<title>{} schema</title>'.format(self.title))
            # gird for battle
            self.f.write('\n</head>\n'
                         '<body>\n')
        # wrap the whole thing in a unique identifier
        self.f.write('<div id="json-schema">')

        # add a little header
        if self.title:
            self.f.write('<h1>{}</h1>\n'.format(self.title))

    def done(self):
        self.f.write('</div>')
        self.f.write('</body>')
        self.f.write('</html>')
        self.f.close()

    def writeln(self, text):
        self.f.write('  ' * self._indent + text + '\n')

    def begin_object(self, name):
        self.writeln('<div class="object">')
        if name:
            self.writeln('<span class="name">{}<span>'.format(name))
        self._indent += 1

    def end_object(self):
        self._indent -= 1
        self.writeln('</div>')

    def begin_array(self, name):
        self.writeln('<div class="array">')
        self.writeln('<span class="name">{}</span>'.format(name))

    def end_array(self):
        self.writeln('</div>')

    def begin_child(self):
        self.writeln('<li>')
        self._indent += 1

    def end_child(self):
        self._indent -= 1
        self.writeln('</li>')

    def begin_children(self):
        self.writeln('<ul>')
        self._indent += 1

    def end_children(self):
        self._indent -= 1
        self.writeln('</ul>')

    def begin_array_items(self):
        self.writeln('<div class="array_items">'
                     '<span class="listof">List of:</span>')
        self._indent += 1

    def end_array_items(self):
        self._indent -= 1
        self.writeln('</div>')

    def begin_node(self, name):
        self.writeln('<div class="scalar">')
        if name:
            self.writeln('<span class="name">{}<span>'.format(name))
        self._indent += 1

    def end_node(self):
        self._indent -= 1
        self.writeln('</div>')

    def description(self, text):
        self.writeln('<span class="desc">{}</span>'.format(text))

    def typename(self, name):
        self.writeln('<span class="type">{}</span>'.format(name))

    def examples(self, text_list):
        plural = 's' if len(text_list) > 1 else ''
        self.writeln('<div class="examples">Example{s}:<ul>'
                     .format(s=plural))
        for text in text_list:
            self.writeln('<li>{}</li>'.format(text))
        self.writeln('</ul></div>')

    def enum(self, text_list):
        self.writeln('<div class="enum">Possible values:<ul>')
        for text in text_list:
            self.writeln('<li>{}</li>'.format(text))
        self.writeln('</ul></div>')

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


def main():
    p = argparse.ArgumentParser()
    p.add_argument('schema', help='Schema file', metavar='FILE')
    p.add_argument('-o', '--output', dest='html', default=None,
                   help='HTML output file', metavar='FILE')
    p.add_argument('-e', '--embed', dest='embed', action='store_true',
                   help='Make output HTML suitable for embedding')
    p.add_argument('-s', '--style', dest='css', default=None,
                   help='CSS stylesheet', metavar='FILE')
    p.add_argument('-t', '--title', dest='title', default='',
                   help='Title (default=none)')
    args = p.parse_args()
    output_file = args.html if args.html else sys.stdout
    html_output = HtmlSchemaEvents(output_file, args.css, args.title,
                                   embeddable=args.embed)
    schema_proc = SchemaProcessor(args.schema, html_output)
    status = schema_proc.process()
    return status

if __name__ == '__main__':
    sys.exit(main())
