##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
import numpy as np
from numpy.linalg import inv
from numpy import linalg as LA
import math as mt
import matplotlib.pyplot as plt
import pandas as pd


def ellipsoid(Cov,x1,x2,PE=False,title=""):
    '''
    Arguments
    Cov = covariance matrix(reduced hessian matrix)
    x1 = point estimate of parameter 1
    x2 = point estimate of parameter 2
    PE = default is False. This plots ellipsoid about the origin
         if True ellipsoid will be plotted about the point estimate
    
    Returns a tuple(0,1,2,3)
    0 =index for parameter 1
    1= index for parameter 2
    2= error ellipse center at the origin
    3= error ellipse center at the point estimates
    title = title of ellipsoid plot
    '''
    title = " 95% confidence ellipsoid :  " + title

    #get the eig matrix turple
    #eig_matrix =([eig_values],[eig_vectors])
    eig_matrix =LA.eig(Cov)
    #get the index for the largest eig value
    largest_eig_value_index= np.argmax(eig_matrix[0])
    #get index for the smallest eig value
    smallest_eig_value_index= np.argmin(eig_matrix[0])

    #use the index to assign the major and minor axis
    if  largest_eig_value_index == 0:
        x=x1
        xlab =r'$\theta_{1}$'
        y=x2
        ylab=r'$\theta_{2}$'
    else:
        x=x2
        xlab =r'$\theta_{2}$'
        y=x1
        ylab =r'$\theta_{1}$'

    #use these indices to get the largest and smallest  eig vectors
    largest_eig_vec=eig_matrix[1][largest_eig_value_index]
    smallest_eig_vec=eig_matrix[1][smallest_eig_value_index]
    # get the largest and smallest eig values too
    largest_eig_value= max(eig_matrix[0])
    smallest_eig_value= min(eig_matrix[0])

    # configure x to represent the major axis(i.e largest eig value)
    #cal angle between the x axis and the largest eig value
    if (abs(largest_eig_vec[0]) < abs(largest_eig_vec[1])):
        angle =mt.atan2(largest_eig_vec[0],largest_eig_vec[1])
    else:
        angle =mt.atan2(largest_eig_vec[1],largest_eig_vec[0])

    while angle < 0 or angle >2* mt.pi:
        if angle <0:
            angle =angle+ 2*mt.pi
        elif angle>2*mt.pi:
            angle=angle -2*mt.pi

    #get 95% confidence interval s=5.991 for df=2
    chi_sq_val =2.4477
    theta_grid = np.linspace(0,2*mt.pi)
    a = chi_sq_val*np.sqrt(largest_eig_value)
    b =chi_sq_val*np.sqrt(smallest_eig_value)

    # the ellipse in x and y coordinate
    phi =angle
    x_r =np.cos(theta_grid)
    y_r =np.sin(theta_grid)
    ellipse_x_r=a*x_r
    ellipse_y_r=b*y_r

    # define a rotation matrix
    R =np.array([[mt.cos(phi),mt.sin(phi)],[-1*mt.sin(phi),mt.cos(phi)]])

    #rotate that ellipse according to variance data
    x_el=np.array([[ellipse_x_r],[ellipse_y_r]])
    r_ellipse = x_el.T
    rotate_ellipse = np.dot(r_ellipse,R)

    #ellisoids for ploting by adding the mean coordinates
    rows =len(rotate_ellipse[...,0])
    cordinates = [x,y]*rows
    cord_mat = np.array(cordinates).reshape(rows,2)
    error_ellipse = rotate_ellipse + cord_mat
    
    #graph
    plt.figure(num=1,figsize=(15,9))
    if PE:
        plt.plot(error_ellipse[...,0],error_ellipse[...,1],c='b')
        plt.scatter(x,y,label='Point estimate',marker='o',c='r')
        plt.legend(fontsize=10)
    else:
        plt.plot(rotate_ellipse[...,0],rotate_ellipse[...,1],c='b')
        
    plt.xlabel(xlab,fontsize=21)
    plt.ylabel(ylab,fontsize=21)
    plt.legend(fontsize=18)
    plt.grid(True, which='major',linestyle='--',axis='x')
    plt.tick_params(labelsize=18)
    plt.title(title,fontsize=21,fontweight='bold')
    plt.tight_layout()
    plt.show()
    return x,y,rotate_ellipse,error_ellipse

if __name__ == "__main__":
    
    #generate the covariance 
    RH_00= 1.7528735631946887e-02
    RH_10=-8.6206896550602399e-04
    RH_11=5.5617352613455781e-05
    RH_01=RH_10
    Cov1 =np.array([[RH_00,RH_01],[RH_10,RH_11]])
    #get point estimates
    P1=4.1775
    P2=0.4871
    #plot ellipsoid
    ellipsoid(Cov1,P1,P2)
    ellipsoid(Cov1,P1,P2,PE =True)
       
                 


                
