##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
This is a simple example or using a Pyomo external to call the IAPWS95 steam
property calculations. The steam properties are implimented as an ASL
(AMPL solver library) user-defined function.

"""
from __future__ import absolute_import  # disable implicit relative imports
from __future__ import division  # No integer division
from __future__ import print_function  # Python 3 style print

_author__ = "John Eslick <john.eslick@netl.doe.gov>"

from pyomo.environ import *
from pyomo.opt import SolverFactory
import idaes.contrib.properties.iapws95 as iapws95
from idaes.core.util.misc import smooth_max
import os

def main(T=275, P=101.325, Fin=1.0, Q=500, hin=112.654899654,
         fix=('T', 'P'), do_opt=True, T_opt=900, P_opt=30000):
    """Test flash with IAPWS properties

    This function is a simple test of a flash using the IAPWS-95 properties for
    steam. It solves a square problem and optionally an optimztion problem. The
    arguments setup the problem.

    Args:
        T: Temperatue of flash (K) or initial guess if not fixed
        P: Pressure of flash (kPA) or intial guess if not fixed
        Q: Heat into flash (kJ/s) or intial guess if not fixed
        Fin: Inlet flowrate (kg/s)
        hin: Inlet mixture specific enthalpy (kJ/kg)
        fix: Variables to fix for square problem, two of {T, P, Q}
        do_opt: True to solve optimzation problem after square, False square only
        T_opt: Temperature target for optimization (K)
        P_opt: Pressure target for optimization (kPa)
    """

    # Get the localtion of the so library
    flib = os.path.dirname(iapws95.__file__)
    flib = os.path.join(flib, "iapws95.so")

    # Set up a very simple Pyomo model
    model = ConcreteModel()

    # Model parameters
    model.eps = Param(initialize=1e-5, doc="Smooth max parameter")
    model.rho_c = Param(initialize=322, doc="Critical density (kg/m3)")
    model.T_c = Param(initialize=647.096, doc="Critical temperature (K)")
    model.T_opt = Param(initialize=T_opt)
    model.P_opt = Param(initialize=P_opt)

    # Load external functions
    model.p = ExternalFunction(library=flib, function="p",
        doc="pressure in KPa as a function of density/critical density and "
        "critical temperature/temperature")
    model.psat = ExternalFunction(library=flib, function="p_sat",
        doc="saturation pressure in KPa as a function of"
        "critical temperature/temperature")
    model.h = ExternalFunction(library=flib, function="h",
        doc="enthalpy in kJ/kg as a function of density/critical density and "
        "critical temperature/temperature")

    # Model Variables
    model.P = Var(initialize=P, doc="Pressure kPa", bounds=(1, 1e6))
    model.rhoL = Var(domain=NonNegativeReals, initialize=1000, doc="Density kg/m3")
    model.rhoV = Var(domain=NonNegativeReals, initialize=0.003, doc="Density kg/m3")
    model.T = Var(initialize=T, doc="Temperature K", bounds=(270, 5000))
    model.Fin = Var(domain=NonNegativeReals, initialize=Fin, doc="Flow kg/s")
    model.FV = Var(domain=NonNegativeReals, initialize=1, doc="Flow kg/s")
    model.FL = Var(domain=NonNegativeReals, initialize=0, doc="Flow kg/s")
    model.hin = Var(initialize=hin)
    model.Q = Var(initialize=Q, doc="Heat flow kJ/s")

    model.Fin.fix()
    model.hin.fix()
    for i in fix:
        getattr(model, i).fix()

    model.tau = Expression(expr=model.T_c/model.T, doc="Reciprocal reduced T")
    model.T_delta = smooth_max(0, model.T - 644, model.eps)
    model.Psat = Expression(expr=model.psat(model.T_c/(model.T-model.T_delta)))
    model.PL_delta = smooth_max(0, model.Psat - model.P, model.eps)
    model.PV_delta = smooth_max(0, model.P - model.Psat, model.eps)
    model.eq_mas_bal = Constraint(expr=model.Fin == model.FV + model.FL)
    model.hL = Expression(expr=model.h(model.rhoL/model.rho_c, model.tau))
    model.hV = Expression(expr=model.h(model.rhoV/model.rho_c, model.tau))
    model.eq_rhoL = Constraint(expr=model.P + model.PL_delta ==
        model.p(model.rhoL/model.rho_c, model.tau))
    model.eq_rhoV = Constraint(expr=model.P - model.PV_delta ==
        model.p(model.rhoV/model.rho_c, model.tau))

    model.eq_energy_bal = Constraint(expr=
        model.Fin*model.hin + model.Q == model.FV*model.hV + model.FL*model.hL)
    model.c0_eq = Constraint(expr=
        model.FV*model.PV_delta - model.FL*model.PL_delta == 0)
    model.c1_ineq = Constraint(expr=
        model.FV*model.PV_delta >= model.FL*model.PL_delta)
    model.c2_ineq = Constraint(expr=
        model.FV*model.PV_delta <= model.FL*model.PL_delta)
    model.c1_ineq.deactivate()
    model.c2_ineq.deactivate()

    # Solve Square Problem
    strip_bounds = TransformationFactory('contrib.strip_var_bounds')
    strip_bounds.apply_to(model, reversible=True)
    opt = SolverFactory('ipopt')
    res = opt.solve(model, tee=True)
    strip_bounds.revert(model)
    display(model)

    # Display some information
    print("\n**Square Problem**\n")
    print("Saturation Pressure = {} kPa".format(value(model.psat(model.tau))))
    print("Liquid Enthalpy = {} kJ/kg".format(value(model.hL)))
    print("Vapor Enthalpy = {} kJ/kg".format(value(model.hV)))
    print("Mixture Enthalpy = {} kJ/kg".format(
        value((model.hV*model.FV + model.hL*model.FL)/(model.FV+model.FL))))
    print("")

    # Maybe do optimization
    if not do_opt:
        return

    # Optimzation Problem
    model.obj = Objective(expr=(model.T-model.T_opt)**2 + (model.P-model.P_opt)**2)
    model.c0_eq.deactivate()
    model.c1_ineq.activate()
    model.c2_ineq.activate()
    model.T.unfix()
    model.Q.unfix()
    model.P.unfix()

    # Solve Optimization Problem
    opt = SolverFactory('ipopt')
    res = opt.solve(model, tee=True)
    display(model)

if __name__ == "__main__":
    main()
