##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
This is a simple example or using a Pyomo external to call the IAPWS95 steam
property calculations. The steam properties are implimented as an ASL
(AMPL solver library) user-defined function.

"""
from __future__ import absolute_import  # disable implicit relative imports
from __future__ import division  # No integer division
from __future__ import print_function  # Python 3 style print

_author__ = "John Eslick <john.eslick@netl.doe.gov>"

from pyomo.environ import *
from pyomo.opt import SolverFactory
import idaes.contrib.properties.iapws95.iapws95_wrap as iapws95
from idaes.core.util.misc import smooth_max
import os

def main(T=500, P=101325, Fin=1.0/18.01528, Q=500, hin=112654.899654*18.01528):
    """Test flash with IAPWS properties
    """
    # Set up a very simple Pyomo model
    model = ConcreteModel()
    model.prop_param = iapws95.PropertyParameterBlock()
    model.prop = iapws95.PropertyBlock(parameters=model.prop_param)

    # Model Variables
    model.Fin = Var(domain=NonNegativeReals, initialize=Fin, doc="Flow kmol/s")
    model.FV = Var(domain=NonNegativeReals, initialize=1, doc="Flow kmol/s")
    model.FL = Var(domain=NonNegativeReals, initialize=0, doc="Flow kmol/s")
    model.hin = Var(initialize=hin)
    model.Q = Var(initialize=Q, doc="Heat flow J/s")

    model.Fin.fix(Fin)
    model.hin.fix()
    model.prop.temperature.fix(T)
    model.prop.pressure.fix(P)

    model.eq_mas_bal = Constraint(expr=model.Fin == model.FV + model.FL)
    model.eq_energy_bal = Constraint(expr=model.Fin*model.hin + model.Q ==
        model.FV*model.prop.enth_vap + model.FL*model.prop.enth_liq)

    model.c0_eq = Constraint(expr=
        model.FV*model.prop.PV_delta - model.FL*model.prop.PL_delta == 0)

    # Solve Square Problem
    strip_bounds = TransformationFactory('contrib.strip_var_bounds')
    strip_bounds.apply_to(model, reversible=True)
    opt = SolverFactory('ipopt')
    res = opt.solve(model, tee=True)
    strip_bounds.revert(model)
    display(model)

    # Display some information
    print("\n**Square Problem**\n")
    print("Saturation Pressure = {} Pa".format(value(model.prop.saturation_pressure)))
    print("Liquid Enthalpy = {} J/kmol".format(value(model.prop.enth_liq)))
    print("Vapor Enthalpy = {} J/kmol".format(value(model.prop.enth_vap)))
    print("Mixture Enthalpy = {} J/kmol".format(
        value((model.prop.enth_vap*model.FV + model.prop.enth_liq*model.FL)/(model.FV+model.FL))))
    print("")

if __name__ == "__main__":
    main()
