##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
This is a simple example or using a Pyomo external to call the IAPWS95 steam
property calculations. The steam properties are implimented as an ASL
(AMPL solver library) user-defined function.
"""
from __future__ import absolute_import  # disable implicit relative imports
from __future__ import division  # No integer division
from __future__ import print_function  # Python 3 style print

__author__ = "John Eslick <john.eslick@netl.doe.gov>"

from pyomo.environ import *
from pyomo.opt import SolverFactory
import idaes.contrib.properties.iapws95.iapws95_wrap as iapws95
from idaes.core.util.misc import smooth_max
import os

def main(Tin=400, x=0.7, T=300, Fin=1.0):
    """Test flash with IAPWS properties

    This assumes the inlet is super heated, better off just using enthalpy in
    if it may be saturated.

    For this simple model the condenser tempertuature is just fixed, for a more
    detailed model there should be an addtional heat transfer submodel.
    """
    # Set up a very simple Pyomo model
    model = ConcreteModel()
    model.prop_param = iapws95.PropertyParameterBlock()
    model.prop_in = iapws95.PropertyBlock(parameters=model.prop_param)
    model.prop = iapws95.PropertyBlock(parameters=model.prop_param)

    # Model Variables
    model.Fin = Var(initialize=Fin, doc="Total inlet flow")
    model.Q = Var(initialize=-1000, doc="Heat flow J/s")

    model.prop.vapor_fraction.fix(0)
    model.prop_in.vapor_fraction.fix(x)
    model.prop.eq_complimentarity.deactivate()
    #model.prop_in.eq_complimentarity.deactivate()
    model.prop_in.temperature.fix(Tin)
    model.prop.temperature.fix(T)
    model.Fin.fix(Fin)

    model.eq_mass_bal = Constraint(expr=model.Fin ==
        model.prop.flow_component_liq["H2O"])
    model.eq_pressure = Constraint(expr=model.prop.pressure ==
        model.prop.saturation_pressure)
    model.eq_energy_bal = Constraint(expr= model.prop_in.enth*model.Fin +
        model.Q == model.prop.enth*model.Fin)

    # Solve Square Problem
    strip_bounds = TransformationFactory('contrib.strip_var_bounds')
    strip_bounds.apply_to(model, reversible=True)
    opt = SolverFactory('ipopt', options={"bound_push":1e-10, "mu_init":1e-10})
    res = opt.solve(model, tee=True)
    strip_bounds.revert(model)
    display(model)

if __name__ == "__main__":
    main()
