##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
This is a simple example or using a Pyomo external to call the IAPWS95 steam
property calculations. The steam properties are implimented as an ASL
(AMPL solver library) user-defined function.

"""
from __future__ import absolute_import  # disable implicit relative imports
from __future__ import division  # No integer division
from __future__ import print_function  # Python 3 style print

_author__ = "John Eslick <john.eslick@netl.doe.gov>"

from pyomo.environ import *
from pyomo.opt import SolverFactory
import idaes.contrib.properties.iapws95.iapws95_wrap as iapws95

import os

import pandas as pd
import matplotlib.pyplot as plt

def main():
    model = ConcreteModel()
    model.prop_param = iapws95.PropertyParameterBlock()
    model.prop = iapws95.PropertyBlock(parameters=model.prop_param)

    step = 0.1
    start = -200
    stop = 1200
    T = 300
    n = int((stop - start)/step)
    p_func = model.prop.p
    tau = model.prop.tau
    delta = model.prop.delta_liq
    density = [0]*n
    pressure = [0]*n
    model.prop.temperature.value = T
    if T <= 647.096:
        Psat = value(model.prop.psat(tau))
    else:
        Psat = -100
    Pc = model.prop.pressure_crit["H2O"]/1000
    print("Psat = {}kPa".format(Psat))
    for i in range(n):
        model.prop.density_mass_liq.value = start + step*i
        density[i] = start + step*i
        pressure[i] = value(p_func(delta, tau))

    plt.figure(num=1,figsize=(12,6))
    plt.subplot(121)
    plt.plot(density,pressure,color='black')
    #plt.plot([50, 100], [50,100],color='black')
    #plt.axis([50,100,50,100])
    plt.xlabel('Density (kg/m3)',fontsize=15)
    plt.ylabel('Pressure (kPa)',fontsize=15)
    plt.tight_layout()
    plt.tick_params(labelsize=14)
    plt.title('Pressure as a function of density for T = {}K'.format(T),fontsize=14)
    plt.subplot(122)
    plt.plot([-500, 1500], [Psat, Psat],color='blue')
    plt.plot([-500, 1500], [Pc, Pc],color='red')
    plt.plot(density,pressure,color='black')
    plt.axis([start,stop,0,30000])
    plt.xlabel('Density (kg/m3)',fontsize=15)
    plt.ylabel('Pressure (kPa)',fontsize=15)
    plt.tight_layout()
    plt.tick_params(labelsize=14)
    plt.title('Pressure as a function of density for T = {}K'.format(T),fontsize=14)
    plt.show()

if __name__ == '__main__':
    main()
