##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
This is a simple example or using a Pyomo external to call the IAPWS95 steam
property calculations. The steam properties are implimented as an ASL
(AMPL solver library) user-defined function.

"""
from __future__ import absolute_import  # disable implicit relative imports
from __future__ import division  # No integer division
from __future__ import print_function  # Python 3 style print

_author__ = "John Eslick <john.eslick@netl.doe.gov>"

from pyomo.environ import *
from pyomo.opt import SolverFactory
import idaes.contrib.properties.iapws95 as iapws95
import os

def main():
    """ """
    # Get the localtion of the so library
    flib = os.path.dirname(iapws95.__file__)
    flib = os.path.join(flib, "iapws95.so")

    # Set up a very simple Pyomo model
    model = ConcreteModel()

    # Model parameters
    model.eps = Param(initialize=1e-5, doc="Smooth max parameter")
    model.rho_c = Param(initialize=322, doc="Critical density (kg/m3)")
    model.T_c = Param(initialize=647.096, doc="Critical temperature (K)")

    # Load external functions
    model.p = ExternalFunction(library=flib, function="p",
        doc="pressure in KPa as a function of density/critical density and "
        "critical temperature/temperature")
    model.h = ExternalFunction(library=flib, function="h",
        doc="enthalpy in kJ/kg as a function of density/critical density and "
        "critical temperature/temperature")
    model.s = ExternalFunction(library=flib, function="s",
        doc="entropy in kJ/kg/K as a function of density/critical density and "
        "critical temperature/temperature")

    # Model Variables
    model.P1 = Var(initialize=4e2, doc="Inlet Pressure (kPa)", bounds=(1, 1e6))
    model.P2 = Var(initialize=2e2, doc="Outlet Pressure (kPa)", bounds=(1, 1e6))
    model.rho1 = Var(domain=NonNegativeReals,
        initialize=0.003, doc="Inlet density (kg/m3)")
    model.rho2 = Var(domain=NonNegativeReals,
        initialize=0.003, doc="Outlet density (kg/m3)")
    model.rhoisen = Var(domain=NonNegativeReals,
        initialize=0.003, doc="Isentropic outlet Density (kg/m3)")
    model.T1 = Var(initialize=1000,
        doc="Inlet temperature (K)", bounds=(270, 5000))
    model.T2 = Var(initialize= 900,
        doc="Outlet temperature (K)", bounds=(270, 5000))
    model.Tisen = Var(initialize=900,
        doc="Isentroptic outlet temperature (K)", bounds=(270, 5000))
    model.h1 = Var(doc="Inlet specific enthalpy (kJ/kg)")
    model.h2 = Var(doc="Outlet specific enthalpy (kJ/kg)")
    model.hisen = Var(doc="Isentropic outley specific enthalpy (kJ/kg)")
    model.s1 = Var(doc="Inlet specific entropy (kJ/kg/K)")
    model.W = Var(initialize=100, doc="Work (kJ/kg)")
    model.Wisen = Var(initialize=100, doc="Isentropic work (kJ/kg)")
    model.eff = Var(initialize=0.9, doc="Isentropic efficency (W/Wisen)")

    model.eff.fix(0.85)
    model.T1.fix(1000)
    model.P1.fix(4e2)
    model.P2.fix(2e2)

    model.rho1_eq = Constraint(expr=model.P1 ==
        model.p(model.rho1/model.rho_c, model.T_c/model.T1))
    model.s1_eq = Constraint(expr=model.s1 ==
        model.s(model.rho1/model.rho_c, model.T_c/model.T1))
    model.h1_eq = Constraint(expr=model.h1 ==
        model.h(model.rho1/model.rho_c, model.T_c/model.T1))
    model.rhoisen_eq = Constraint(expr=model.P2 ==
        model.p(model.rhoisen/model.rho_c, model.T_c/model.Tisen))
    model.hisen_eq = Constraint(expr=model.hisen ==
        model.h(model.rhoisen/model.rho_c, model.T_c/model.Tisen))
    model.Tisen_eq = Constraint(expr=model.s1 ==
        model.s(model.rhoisen/model.rho_c, model.T_c/model.Tisen))
    model.Wisen_eq = Constraint(expr=model.Wisen == model.h1 - model.hisen)
    model.eff_eq = Constraint(expr=model.eff*model.Wisen == model.W)
    model.energy_eq = Constraint(expr=model.W == model.h1 - model.h2)
    model.T2_eq = Constraint(expr=model.h2 ==
        model.h(model.rho2/model.rho_c, model.T_c/model.T2))
    model.rho2_eq = Constraint(expr=model.P2 ==
        model.p(model.rho2/model.rho_c, model.T_c/model.T2))

    # Solve Square Problem
    strip_bounds = TransformationFactory('contrib.strip_var_bounds')
    strip_bounds.apply_to(model, reversible=True)
    opt = SolverFactory('ipopt')
    res = opt.solve(model, tee=True)
    strip_bounds.revert(model)

    display(model)

if __name__ == "__main__":
    main()
