##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
This is a simple example or using a Pyomo external to call the IAPWS95 steam
property calculations. The steam properties are implimented as an ASL
(AMPL solver library) user-defined function.

"""
from __future__ import absolute_import  # disable implicit relative imports
from __future__ import division  # No integer division
from __future__ import print_function  # Python 3 style print

_author__ = "John Eslick <john.eslick@netl.doe.gov>"

from pyomo.environ import *
from pyomo.opt import SolverFactory
import idaes.contrib.properties.iapws95.iapws95_wrap as iapws95
import os


def main():
    """ """
    # Get the localtion of the so library
    flib = os.path.dirname(iapws95.__file__)
    flib = os.path.join(flib, "iapws95.so")

    # Set up a very simple Pyomo model
    model = ConcreteModel()
    model.prop_param = iapws95.PropertyParameterBlock()
    model.prop1 = iapws95.PropertyBlock(parameters=model.prop_param,
        doc="Inlet properties", calculate_equilibrium=False)
    model.prop2 = iapws95.PropertyBlock(parameters=model.prop_param,
        doc="Outlet properties")
    model.prop_isen = iapws95.PropertyBlock(parameters=model.prop_param,
        doc="Isentropic outlet properties")

    model.W = Var(initialize=-1e5, doc="Work [J/kmol]")
    model.Wisen = Var(initialize=-1e5, doc="Isentropic work [J/kmol]")
    model.eff = Var(initialize=0.9, doc="Isentropic efficency [W/Wisen]")

    # Shorter names
    T1 = model.prop1.temperature
    T2 = model.prop2.temperature
    Tisen = model.prop_isen.temperature
    P1 = model.prop1.pressure
    P2 = model.prop2.pressure
    Pisen = model.prop_isen.pressure
    h1 = model.prop1.enth_mol
    h2 = model.prop2.enth_mol
    hisen = model.prop_isen.enth_mol
    s1 = model.prop1.entr_mol
    sisen = model.prop_isen.entr_mol
    eff = model.eff
    F1 = model.prop1.flow_mol
    F2 = model.prop2.flow_mol
    Fisen = model.prop_isen.flow_mol
    #{'flow_mol': 30322.477759999998, 'pressure': 6142813.975349435, 'temperature': 635.7961601820855, 'vapor_frac': 1.0000000000000002}
    model.prop1.vapor_frac.fix(1.0000000000000002)
    F1.fix(30322.477759999998)
    T1.fix(635.7961601820855)    # K
    P1.fix(6142813.975349435)  # Pa
    P2.fix(6142813.975349435*0.93)       # Pa
    eff.fix(0.90)

    Tisen.fix(626.8)
    #T2.fix(626.8)

    model.mb2_eq = Constraint(expr=F2 == F1)
    model.md2_eq = Constraint(expr=Fisen == F1)
    model.P_eq = Constraint(expr=Pisen == P2)
    #model.s_eq = Constraint(expr= 0 == sisen - s1)
    model.Wisen_eq = Constraint(expr=model.Wisen == h1*F1 - hisen*F1)
    model.W_eq = Constraint(expr=eff*model.Wisen == model.W)
    model.h2_eq = Constraint(expr=model.W == h1*F1 - h2*F1)
    # Solve Square Problem
    strip_bounds = TransformationFactory('contrib.strip_var_bounds')
    strip_bounds.apply_to(model, reversible=True)
    display(model)
    opt = SolverFactory('ipopt')
    res = opt.solve(model, tee=True, options={"bound_push":1e-10})
    #strip_bounds.revert(model)

    display(model)
    print("Pover = {}".format(value(model.prop2.Punder)))
    print("Punder = {}".format(value(model.prop2.Pover)))
    print("Psat = {}".format(value(model.prop2.pressure_sat)))
    print("P = {}".format(value(model.prop2.pressure)))
    print("Sisen = {}".format(value(sisen)))
    print("S1 = {}".format(value(s1)))
    print("H1 = {}".format(value(h1)))
    print("dens_liq = {}".format(value(model.prop2.dens_mass_phase["Liq"])))
    print("dens_vap = {}".format(value(model.prop2.dens_mass_phase["Vap"])))

    #for T in range(540*10,640*10,1):
    #    model.prop_isen.temperature.value = T/10
    #    print("{}   {}   {}".format(T/10, value(sisen), value(hisen)))


if __name__ == "__main__":
    main()
