##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration property package for heterogeneous systems
"""

# Chages the divide behavior to not do integer division
from __future__ import division

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, Expression, exp, log10, Param, \
                          PositiveReals, RangeSet, Reals, Set, value, Var
from pyomo.opt import SolverFactory, TerminationCondition

# Import IDAES cores
from idaes.core import declare_process_block_class, PropertyBlockBase, \
                        PropertyBlockDataBase, PropertyParameterBase
from idaes.core.util.misc import add_object_ref, solve_indexed_blocks

# Some more inforation about this module
__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.unit_model.properties')


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with proeprties for
    a demonstration VLSE system.

    """

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PropertyParameterData, self).build()

        self._make_params()

    def _make_params(self):
        '''
        This section is for parameters needed for the property models.
        '''
        # List of valid phases in property package
        self.phase_list = Set(initialize=['Vap', 'Liq'])

        # Component list - a list of component identifiers
        self.component_list_vap = Set(initialize=['CO2', 'H2O', 'N2'])
        self.component_list_liq = Set(initialize=['CO2', 'H2O'])

        self.component_list = (self.component_list_vap |
                               self.component_list_liq)

        # List of components in each phase (optional)
        self.phase_component_list = {"Vap": self.component_list_vap,
                                     "Liq": self.component_list_liq}

        # List of Reaction Indicies
        self.phase_equilibrium_idx = Set(initialize=[1, 2])

        # Reaction Stoichiometry
        self.phase_equilibrium_list = \
            {1: ["CO2", ("Vap", "Liq")],
             2: ["H2O", ("Vap", "Liq")]}

        # Thermodynamic reference state
        self.temperature_reference = Param(within=PositiveReals,
                                           mutable=True,
                                           default=303.15,
                                           doc='Reference temperature [K]')

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

        # Heat capacities
        self.cp_mol_liq = Param(within=PositiveReals,
                                mutable=True,
                                default=75.3,
                                doc='Liquid phase heat capacity [J/mol.K]')
        self.cp_mol_vap = Param(within=PositiveReals,
                                mutable=True,
                                default=28.0,
                                doc='Vapour phase heat capacity [J/mol.K]')

        # Henry constant coefficients
        iv_H = {1: 630,
                2: 2400}
        self.henry_coeff = Param(RangeSet(2), initialize=iv_H, mutable=True,
                                 doc="Cofficients for Henry constant")

        # Antoine coefficients
        iv_Pvap = {1: 6.21,
                   2: 2357.7,
                   3: 7.56}
        self.antoine = Param(RangeSet(3), initialize=iv_Pvap,
                             mutable=True,
                             doc="Antoine coefficients")

    def get_supported_properties(self):
        return {'flow_mol_phase_comp': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'},
                'dens_mol_phase': {'method': '_dens_mol_phase',
                                   'units': 'mol/m^3'},
                'enth_mol_liq': {'method': '_enth_mol_liq',
                                 'units': 'J/mol'},
                'enth_mol_vap': {'method': '_enth_mol_vap',
                                 'units': 'J/mol'},
                'henry': {'method': '_henry', 'units': 'mol/kg.bar'},
                'mole_frac_liq': {'method': '_mole_frac_liq', 'units': None},
                'mole_frac_vap': {'method': '_mole_frac_vap', 'units': None},
                'pressure_sat': {'method': '_pressure_sat', 'units': 'bar'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


class _PropertyBlock(PropertyBlockBase):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """
    def initialize(blk, flow_mol_phase_comp=None,
                   pressure=None, temperature=None,
                   outlvl=0, optarg=None, solver='ipopt', hold_state=False):
        '''
        Initialisation routine for property package.

        Keyword Arguments:
            flow_mol_phase_comp : value at which to initialize phase
                                  component flows (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature
                          (default=None)
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)

            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        '''
        # Fix state variables if not already fixed
        Fcvflag = {}
        Fclflag = {}
        Pflag = {}
        Tflag = {}

        for k in blk.keys():
            for j in blk[k].component_list_vap:
                if blk[k].flow_mol_phase_comp['Vap', j].fixed is True:
                    Fcvflag[k, j] = True
                else:
                    Fcvflag[k, j] = False
                    if flow_mol_phase_comp is None:
                        blk[k].flow_mol_phase_comp['Vap', j].fix(1.0)
                    else:
                        blk[k].flow_mol_phase_comp['Vap', j].fix(
                                flow_mol_phase_comp['Vap', j])

            for j in blk[k].component_list_liq:
                if blk[k].flow_mol_phase_comp['Liq', j].fixed is True:
                    Fclflag[k, j] = True
                else:
                    Fclflag[k, j] = False
                    if flow_mol_phase_comp is None:
                        blk[k].flow_mol_phase_comp['Liq', j].fix(1.0)
                    else:
                        blk[k].flow_mol_phase_comp['Liq', j].fix(
                                flow_mol_phase_comp['Liq', j])

            if blk[k].pressure.fixed is True:
                Pflag[k] = True
            else:
                Pflag[k] = False
                if pressure is None:
                    blk[k].pressure.fix(101325.0)
                else:
                    blk[k].pressure.fix(pressure)

            if blk[k].temperature.fixed is True:
                Tflag[k] = True
            else:
                Tflag[k] = False
                if temperature is None:
                    blk[k].temperature.fix(300.0)
                else:
                    blk[k].temperature.fix(temperature)

        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        if optarg is None:
            sopt = {'tol': 1e-8}
        else:
            sopt = optarg

        opt = SolverFactory('ipopt')
        opt.options = sopt

        # ---------------------------------------------------------------------
        # Deactivate constraints
        for k in blk.keys():
            if hasattr(blk[k], "henry_H2O"):
                blk[k].henry_H2O.deactivate()
            if hasattr(blk[k], "Antoine_correlation"):
                blk[k].Antoine_correlation.deactivate()

                blk[k].pressure_sat = (10**(value(blk[k].antoine[1]) -
                                       value(blk[k].antoine[2]) /
                                       (value(blk[k].temperature) +
                                        value(blk[k].antoine[3]))))

            for j in blk[k].component_list_liq:
                blk[k].mole_frac_liq[j] = (
                        value(blk[k].flow_mol_phase_comp['Liq', j]) /
                        sum(value(blk[k].flow_mol_phase_comp['Liq', i])
                            for i in blk[k].component_list_liq))

            for j in blk[k].component_list_vap:
                blk[k].mole_frac_vap[j] = (
                        value(blk[k].flow_mol_phase_comp['Vap', j]) /
                        sum(value(blk[k].flow_mol_phase_comp['Vap', i])
                            for i in blk[k].component_list_vap))

        results = solve_indexed_blocks(opt, blk, tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                logger.info('{} Initialisation Step 1 Complete.'
                            .format(blk.name))
            else:
                logger.warning('{} Initialisation Step 1 Failed.'
                               .format(blk.name))

        # ---------------------------------------------------------------------
        # Activate constraints
        for k in blk.keys():
            if hasattr(blk[k], "henry_H2O"):
                blk[k].henry_H2O.activate()
            if hasattr(blk[k], "Antoine_correlation"):
                blk[k].Antoine_correlation.activate()

        # If input block, return flags, else release state
        flags = {"Fcvflag": Fcvflag,
                 "Fclflag": Fclflag,
                 "Pflag": Pflag,
                 "Tflag": Tflag}

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} Initialisation Complete.'.format(blk.name))

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags)

    def release_state(blk, flags, outlvl=0):
        '''Method to relase state variables fixed during initialisation'''
        # Unfix state variables
        for k in blk.keys():
            for j in blk[k].component_list_vap:
                if flags['Fcvflag'][k, j] is False:
                    blk[k].flow_mol_phase_comp['Vap', j].unfix()
            for j in blk[k].component_list_liq:
                if flags['Fclflag'][k, j] is False:
                    blk[k].flow_mol_phase_comp['Liq', j].unfix()
            if flags['Pflag'][k] is False:
                blk[k].pressure.unfix()
            if flags['Tflag'][k] is False:
                blk[k].temperature.unfix()

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} State Released.'.format(blk.name))


@declare_process_block_class("PropertyBlock",
                             block_class=_PropertyBlock)
class PropertyBlockData(PropertyBlockDataBase):
    """
    This is an example of a property package for calcuating the thermophysical
    properties of a demonstration VLSE system.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(PropertyBlockData, self).build()

        self._make_params()
        self._make_vars()
        self._make_constraints()
        self._make_balance_terms()

    def _make_params(self):
        '''
        This section makes references to the necessary parameters contained
        within the Property Parameter Block provided.
        '''
        # List of valid phases in property package
        add_object_ref(self, "phase_list", self.config.parameters.phase_list)

        # Component list - a list of component identifiers
        add_object_ref(self, "component_list_vap",
                       self.config.parameters.component_list_vap)
        add_object_ref(self, "component_list_liq",
                       self.config.parameters.component_list_liq)

        self.component_list = (self.component_list_vap |
                               self.component_list_liq)

        # List of Reaction Indicies
        add_object_ref(self, "phase_equilibrium_idx",
                       self.config.parameters.phase_equilibrium_idx)

        # Reaction Stoichiometry
        add_object_ref(
                self,
                "phase_equilibrium_list",
                self.config.parameters.phase_equilibrium_list)

        # Thermodynamic reference state
        add_object_ref(self, "temperature_reference",
                       self.config.parameters.temperature_reference)

        # Gas Constant
        add_object_ref(self, "gas_const",
                       self.config.parameters.gas_const)

    def _make_vars(self):
        ''' This section contains the state variables to be used to calulcate
            the properties, and needs to be linked to the state properties in
            the unit model.
        '''
        self.flow_mol_phase_comp = Var(self.phase_list,
                                       self.component_list,
                                       domain=Reals,
                                       initialize=0.0,
                                       bounds=(0, 1e3),
                                       doc='Phase Component molar '
                                           'flowrates [mol/s]')
        self.pressure = Var(domain=Reals,
                            initialize=101325.0,
                            bounds=(1e3, 10e5),
                            doc='State pressure [Pa]')
        self.temperature = Var(domain=Reals,
                               initialize=303.15,
                               bounds=(273.15, 373.15),
                               doc='State temperature [K]')

    def _make_constraints(self):
        # Create standard constraints
        ''' This section creates the necessary constraints for calculating
            the property values.
        '''
        if ((self.config.calculate_phase_equilibrium is True) and
                (self.config.has_sum_fractions is True)):
            self.raoult_H2O = Constraint(expr=self.pressure_sat ==
                                         1e-5*self.pressure *
                                         self.mole_frac_vap["H2O"])
            self.henry_H2O = Constraint(
                                expr=self.henry*self.mole_frac_liq["CO2"] ==
                                self.mole_frac_vap["CO2"]*self.pressure*1e-5)

    def _dens_mol_phase(self):
        self.dens_mol_phase = Var(self.phase_list, doc="Molar density ")

        def dens_mol_phase_calculation(b, p):
            if p == "Vap":
                return b.pressure == (b.dens_mol_phase[p]*b.gas_const *
                                      b.temperature)
            elif p == "Liq":
                return b.dens_mol_phase[p] == 55555.0
        try:
            # Try to build constraint
            self.dens_mol_phase_calculation = Constraint(
                        self.phase_list,
                        rule=dens_mol_phase_calculation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.dens_mol_phase)
            self.del_component(self.dens_mol_phase_calculation)
            raise

    def _enth_mol_liq(self):
        # Liquid phase enthalpy
        self.enth_mol_liq = Var()
        add_object_ref(self, "cp_mol_liq",
                       self.config.parameters.cp_mol_liq)
        self.enth_mol_liq_correlation = Constraint(
                expr=self.enth_mol_liq ==
                self.cp_mol_liq *
                (self.temperature-self.temperature_reference))

    def _enth_mol_vap(self):
        # Vapour phase enthalpy
        self.enth_mol_vap = Var()
        add_object_ref(self, "cp_mol_vap",
                       self.config.parameters.cp_mol_vap)
        self.enth_mol_vap_correlation = Constraint(
                expr=self.enth_mol_vap ==
                self.cp_mol_vap *
                (self.temperature-self.temperature_reference))

    def _henry(self):
        self.henry = Var()  # mol/kg.bar
        add_object_ref(self, "henry_coeff",
                       self.config.parameters.henry_coeff)
        self.eq_henry = Constraint(expr=self.henry ==
                                   self.henry_coeff[1] *
                                   exp(self.henry_coeff[2] *
                                       (self.temperature**-1 -
                                        self.temperature_reference**-1)))

    def _pressure_sat(self):
        # Vapour pressure
        self.pressure_sat = Var(initialize=1, bounds=(1e-5, 1e2))  # bar
        add_object_ref(self, "antoine",
                       self.config.parameters.antoine)
        A = self.antoine
        self.Antoine_correlation = Constraint(
                expr=(A[1] - log10(self.pressure_sat)) *
                (self.temperature+A[3]) == A[2])

    def _mole_frac_liq(self):
        # Liquid phase mole fractions
        self.mole_frac_liq = Var(self.component_list_liq)

        def mole_frac_liq_calculation(b, j):
            return self.flow_mol_phase_comp['Liq', j] == \
                    (self.mole_frac_liq[j] *
                     sum(self.flow_mol_phase_comp['Liq', k]
                         for k in self.component_list_liq))
        self.mole_frac_liq_calculation = Constraint(
                            self.component_list_liq,
                            rule=mole_frac_liq_calculation)

    def _mole_frac_vap(self):
        # Vapour phase mole fractions
        self.mole_frac_vap = Var(self.component_list_vap)

        def mole_frac_vap_calculation(b, j):
            return self.flow_mol_phase_comp['Vap', j] == \
                    (self.mole_frac_vap[j] *
                     sum(self.flow_mol_phase_comp['Vap', k]
                         for k in self.component_list_vap))
        self.mole_frac_vap_calculation = Constraint(
                            self.component_list_vap,
                            rule=mole_frac_vap_calculation)

    def _make_balance_terms(self):
        def material_terms(b, i, j):
            return b.flow_mol_phase_comp[i, j]
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, i):
            if i == "Vap":
                return (sum(b.flow_mol_phase_comp['Vap', k]
                            for k in self.component_list_vap)*b.enth_mol_vap)
            elif i == "Liq":
                return (sum(b.flow_mol_phase_comp['Liq', k]
                            for k in self.component_list_liq)*b.enth_mol_liq)
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            if p == "Liq":
                if j in b.component_list_liq:
                    return b.dens_mol_phase[p]*b.mole_frac_liq[j]
                else:
                    return 0
            elif p == "Vap":
                if j in b.component_list_vap:
                    return b.dens_mol_phase[p]*b.mole_frac_vap[j]
                else:
                    return 0
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            if p == "Liq":
                return b.dens_mol_phase[p]*b.enth_mol_liq
            elif p == "Vap":
                return b.dens_mol_phase[p]*b.enth_mol_vap
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"flow_mol_phase_comp": b.flow_mol_phase_comp,
                   "temperature": b.temperature,
                   "pressure": b.pressure}
        return members

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            logger.error('{} Temperature set below lower bound.'
                         .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            logger.error('{} Temperature set above upper bound.'
                         .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            logger.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            logger.error('{} Pressure set above upper bound.'.format(blk.name))
