##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Basic property package for superheated steam
"""

# Chages the divide behavior to not do integer division
from __future__ import division

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, Expression, exp, log, Param, \
                          PositiveReals, RangeSet, Reals, Set, value, Var
from pyomo.opt import SolverFactory, TerminationCondition

# Import IDAES cores
from idaes.core import declare_process_block_class, PropertyBlockBase, \
                        PropertyBlockDataBase, PropertyParameterBase
from idaes.core.util.misc import add_object_ref, solve_indexed_blocks

# Some more inforation about this module
__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.unit_model.properties')


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PropertyParameterData, self).build()

        self._make_params()

    def _make_params(self):
        '''
        This section is for parameters needed for the property models.
        '''
        # List of valid phases in property package
        self.phase_list = Set(initialize=['Vap'])

        # Component list - a Set of component identifiers
        self.component_list = Set(initialize=['H2O'])

        # List of components in each phase (optional)
        self.phase_component_list = {"Vap": self.component_list}

        # Thermodynamic reference state
        self.pressure_ref = Param(within=PositiveReals,
                                  mutable=True,
                                  default=22.1e6,
                                  doc='Reference pressure [Pa]')
        self.temperature_ref = Param(within=PositiveReals,
                                     mutable=True,
                                     default=298.16,
                                     doc='Reference temperature [K]')

        # Critical Properties
        self.pressure_crit = Param(within=PositiveReals,
                                   mutable=False,
                                   default=22.1e6,
                                   doc='Critical pressure [Pa]')
        self.temperature_crit = Param(within=PositiveReals,
                                      mutable=False,
                                      default=647,
                                      doc='Critical temperature [K]')

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

        # Specific Enthalpy Coefficients
        iv_h = {1: 31814.1054800032,
                2: 34183.7862803747,
                3: -99377.9449208451,
                4: 152436.959011993,
                5: -12906.706688322,
                6: -17016.8781121182,
                7: 4996.85844021475,
                8: -60388.2385329371,
                9: 13553.5957893638,
                10: -958.111095243271}
        self.enth_coeff = Param(RangeSet(10),
                                initialize=iv_h,
                                mutable=True,
                                doc="Coefficients for specific enthalpy")

        # Specfic Entropy Coefficients
        iv_s = {1: 2917.8849287224,
                2: 5563.8788563665,
                3: 2476.76397435096,
                4: -2851.0896664505,
                5: -7710.07346403133,
                6: 294.934968328149,
                7: -663.957642514899,
                8: 371.895820249943,
                9: -299.193020184272,
                10: -9.42439664096366,
                11: 783.248552127974,
                12: 6567.00337543338,
                13: 1613.94047633751,
                14: 4533.66650271276,
                15: 388.665783988954,
                16: 514.861016344497,
                17: -7.74712651304791,
                18: -0.991243233600187,
                19: 3.9830103289155,
                20: 0.416081013802805,
                21: 9.08778000301356,
                22: -3.32971915225717}
        self.entr_coeff = Param(RangeSet(22),
                                initialize=iv_s,
                                mutable=True,
                                doc="Coefficients for specific entropy")

        # Compressibility Coefficients
        iv_z = {1: -0.0100571153727327,
                2: -5.3542757214324,
                3: -0.00478895702459818,
                4: 8.40366401646245,
                5: 0.419630485063482,
                6: -0.0146604228508421,
                7: 2.48477350930031}
        self.compress_fact_coeff = Param(RangeSet(7),
                                         initialize=iv_z,
                                         mutable=True,
                                         doc="Coefficients for vapor "
                                         "compressibility")

        # Heat capacity ratio coefficients
        iv_rc = {1: 0.0487942478239514,
                 2: 19.498413154935,
                 3: 0.0277446397242404,
                 4: -35.4023136910921,
                 5: -1.59388770336934,
                 6: 0.0563418962280028,
                 7: -0.133400431874117,
                 8: -2.79724889525636}
        self.heat_capacity_ratio_coeff = Param(RangeSet(8),
                                               initialize=iv_rc,
                                               mutable=True,
                                               doc="Coefficients for heat "
                                               "capacity ratio")

        # Vapor pressure coefficients
        iv_pvap = {1: 137.41205,
                   2: -8825.25,
                   3: 0.027614,
                   4: -18.8516,
                   5: -6.49694e-6}
        self.p_sat_coeff = Param(RangeSet(8),
                                 initialize=iv_pvap,
                                 mutable=True,
                                 doc="Coefficients for vapor pressure")

    def get_supported_properties(self):
        return {'flow_mol_comp': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'},
                'enth_mol': {'method': None, 'units': 'J/mol'},
                'pressure_red': {'method': None, 'units': None},
                'temperature_red': {'method': None, 'units': None},
                'compress_fact': {'method': '_compress_fact', 'units': None},
                'entr_mol': {'method': '_entr_mol', 'units': 'J/mol.K'},
                'heat_capacity_ratio': {'method': '_heat_capacity_ratio',
                                        'units': None},
                'dens_mol_phase': {'method': '_dens_mol_phase',
                                   'units': 'mol/m^3'},
                'p_sat': {'method': '_p_sat', 'units': 'Pa'},
                'flow_mol': {'method': '_flow_mol', 'units': 'mol/s'},
                'flow_vol': {'method': '_flow_vol', 'units': 'm^3/s'},
                'mw': {'method': '_mw', 'units': 'g/mol'},
                'vap_frac': {'method': 'vap_frac', 'units': None}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


class _PropertyBlock(PropertyBlockBase):
    """
    This Class contains methods which should be applied to a Property Blocks as
    a whole, rather than individual elements of indexed Property Blocks.
    """
    def initialize(blk, flow_mol_comp={"H2O": 27.5e3}, pressure=2.97e7,
                   temperature=866.5, hold_state=False, outlvl=0,
                   solver='ipopt', optarg={'tol': 1e-8}):
        '''
        Initialisation routine for property package.

        Keyword Arguments:
            flow_mol_comp : value at which to initialize component flows
                             (default=27.5e3 mol/s)
            pressure : value at which to initialize pressure
                        (default=2.97e7 Pa)
            temperature : value at which to initialize temperature
                          (default=866.5 K)
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)

            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        '''
        # Fix state variables if not already fixed
        Fcflag = {}
        Pflag = {}
        Tflag = {}

        for k in blk.keys():
            for j in blk[k].component_list:
                if blk[k].flow_mol_comp[j].fixed is True:
                    Fcflag[k, j] = True
                else:
                    Fcflag[k, j] = False
                    if flow_mol_comp is None:
                        blk[k].flow_mol_comp[j].fix(1.0)
                    else:
                        blk[k].flow_mol_comp[j].fix(flow_mol_comp[j])

            if blk[k].pressure.fixed is True:
                Pflag[k] = True
            else:
                Pflag[k] = False
                if pressure is None:
                    blk[k].pressure.fix(101325.0)
                else:
                    blk[k].pressure.fix(pressure)

            if blk[k].temperature.fixed is True:
                Tflag[k] = True
            else:
                Tflag[k] = False
                if temperature is None:
                    blk[k].temperature.fix(300.0)
                else:
                    blk[k].temperature.fix(temperature)

        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        opt = SolverFactory(solver)
        opt.options = optarg

        # ---------------------------------------------------------------------
        # Solve 1st stage
        for k in blk.keys():
            if hasattr(blk[k], "p_sat_correlation"):
                blk[k].p_sat = \
                                exp(blk[k].p_sat_coeff[1].value +
                                    blk[k].p_sat_coeff[2].value /
                                    value(blk.temperature) +
                                    blk[k].p_sat_coeff[3].value *
                                    value(blk.temperature) +
                                    blk[k].p_sat_coeff[4].value *
                                    log(value(blk.temperature)) +
                                    blk[k].p_sat_coeff[5].value *
                                    value(blk.temperature)**2)

            blk[k].enth_mol_correlation.deactivate()
            if hasattr(blk[k], "volumetric_flow_calculation"):
                blk[k].volumetric_flow_calculation.deactivate()
            if hasattr(blk[k], "entropy_correlation"):
                blk[k].entropy_correlation.deactivate()
            if hasattr(blk[k], "dens_mol_phase_calculation"):
                blk[k].dens_mol_phase_calculation.deactivate()

        results = solve_indexed_blocks(opt, blk, tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                logger.info('{} Initialisation Step 1 Complete.'
                            .format(blk.name))
            else:
                logger.warning('{} Initialisation Step 1 Failed.'
                               .format(blk.name))

        # ---------------------------------------------------------------------
        # Solve 2nd stage
        for k in blk.keys():
            blk[k].enth_mol_correlation.activate()
            if hasattr(blk[k], "volumetric_flow_calculation"):
                blk[k].volumetric_flow_calculation.activate()
            if hasattr(blk[k], "entropy_correlation"):
                blk[k].entropy_correlation.activate()
            if hasattr(blk[k], "dens_mol_phase_calculation"):
                blk[k].dens_mol_phase_calculation.activate()

        results = solve_indexed_blocks(opt, blk, tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                logger.info('{} Initialisation Step 2 Complete.'
                            .format(blk.name))
            else:
                logger.warning('{} Initialisation Step 2 Failed.'
                               .format(blk.name))

        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        flags = {"Fcflag": Fcflag, "Pflag": Pflag,
                 "Tflag": Tflag}

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} Initialisation Complete.'.format(blk.name))

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags)

    def release_state(blk, flags, outlvl=0):
        '''
        Method to relase state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        '''
        # Unfix state variables
        for k in blk.keys():
            for j in blk[k].component_list:
                if flags['Fcflag'][k, j] is False:
                    blk[k].flow_mol_comp[j].unfix()
            if flags['Pflag'][k] is False:
                blk[k].pressure.unfix()
            if flags['Tflag'][k] is False:
                blk[k].temperature.unfix()

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} State Released.'.format(blk.name))


@declare_process_block_class("PropertyBlock",
                             block_class=_PropertyBlock)
class PropertyBlockData(PropertyBlockDataBase):
    """
    This is an example of a property package for calculating the thermophysical
    properties of superheated steam.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(PropertyBlockData, self).build()

        self._make_params()
        self._make_state_vars()
        self._make_prop_vars()
        self._make_constraints()
        self._make_balance_terms()

    def _make_params(self):
        '''
        This section makes references to the necessary parameters contained
        within the Property Parameter Block provided.
        '''
        # List of valid phases in property package
        add_object_ref(self, "phase_list", self.config.parameters.phase_list)

        # Component list - a list of component identifiers
        add_object_ref(self, "component_list",
                       self.config.parameters.component_list)

        # Thermodynamic reference state
        add_object_ref(self, "pressure_ref",
                       self.config.parameters.pressure_ref)
        add_object_ref(self, "temperature_ref",
                       self.config.parameters.temperature_ref)

        # Gas Constant
        add_object_ref(self, "gas_const",
                       self.config.parameters.gas_const)

        # Critical Properties
        add_object_ref(self, "pressure_crit",
                       self.config.parameters.pressure_crit)
        add_object_ref(self, "temperature_crit",
                       self.config.parameters.temperature_crit)

        # Specific Enthalpy Coefficients
        add_object_ref(self, "enth_coeff",
                       self.config.parameters.enth_coeff)

    def _make_state_vars(self):
        """
        This section makes the necessary state variable objects
        """
        self.flow_mol_comp = Var(self.component_list,
                                 domain=Reals,
                                 initialize=0.0,
                                 bounds=(0, 1e6),
                                 doc='Component molar flowrate [mol/s]')
        self.pressure = Var(domain=Reals,
                            initialize=23e6,
                            bounds=(1, 5e7),
                            doc='State pressure [Pa]')
        self.temperature = Var(domain=Reals,
                               initialize=647,
                               bounds=(200, 1000),
                               doc='State temperature [K]')

    def _make_prop_vars(self):
        """
        Make additional variables for calculations
        """
        self.enth_mol = Var(doc='Specific Enthalpy [J/mol]')

        self.pressure_red = Var(initialize=1.0, doc='Reduced Pressure')
        self.temperature_red = Var(initialize=1.0, doc='Reduced Temperature')

    def _make_constraints(self):
        '''
        Create property constraints
        '''
        # Phase Specific Enthalpy
        def enth_mol_correlation(b):
            return b.enth_mol == \
                            (b.enth_coeff[1] +
                             b.enth_coeff[2]*b.temperature_red +
                             b.enth_coeff[3]*b.pressure_red +
                             b.enth_coeff[4]*b.temperature_red *
                             b.pressure_red +
                             b.enth_coeff[5]*b.temperature_red**2 +
                             b.enth_coeff[6]*b.pressure_red**2 +
                             b.enth_coeff[7]*b.temperature_red**3 +
                             b.enth_coeff[8]*b.temperature_red**2 *
                             b.pressure_red +
                             b.enth_coeff[9]*b.temperature_red *
                             b.pressure_red**2 +
                             b.enth_coeff[10]*b.pressure_red**3)
        self.enth_mol_correlation = Constraint(rule=enth_mol_correlation)

        # Reduced Temperature and Pressure
        def reduced_pressure_calculation(b):
            return b.pressure_red*b.pressure_crit == b.pressure
        self.reduced_pressure_calculation = Constraint(
                                            rule=reduced_pressure_calculation)

        def reduced_temperature_calculation(b):
            return b.temperature_red*b.temperature_crit == \
                            b.temperature
        self.reduced_temperature_calculation = Constraint(
                                        rule=reduced_temperature_calculation)

    def _compress_fact(self):
        # Compressibility (from Spirax-Sarco via ALAMO)
        self.compress_fact = Var(initialize=0.73,
                                 doc='Compressibility Factor')

        # Add reference to coefficients
        add_object_ref(self, "compress_fact_coeff",
                       self.config.parameters.compress_fact_coeff)

        def compress_fact_correlation(b):
            return b.compress_fact == \
                        (b.compress_fact_coeff[1]*b.pressure*1e-6 +
                         b.compress_fact_coeff[2]*b.temperature/100 +
                         b.compress_fact_coeff[3]*log(b.pressure*1e-6) +
                         b.compress_fact_coeff[4]*log(b.temperature/100) +
                         b.compress_fact_coeff[5]*(b.temperature/100)**2 +
                         b.compress_fact_coeff[6]*(b.temperature/100)**3 +
                         b.compress_fact_coeff[7]*(b.temperature/100)**0.5)
        try:
            # Try to build constraint
            self.compress_fact_correlation = Constraint(
                                        rule=compress_fact_correlation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.compress_fact)
            del self.compress_fact_coeff
            self.del_component(self.compress_fact_correlation)
            raise

    def _entr_mol(self):
        # Specific Entropy
        self.entr_mol = Var(doc='Specific Entropy [J/mol.K]')

        # Add reference to coefficients
        add_object_ref(self, "entr_coeff",
                       self.config.parameters.entr_coeff)

        def entropy_correlation(b):
            return b.entr_mol == \
                        (b.entr_coeff[1] +
                         b.entr_coeff[2]*b.temperature_red +
                         b.entr_coeff[3]*b.pressure_red +
                         b.entr_coeff[4]*b.temperature_red *
                         b.pressure_red +
                         b.entr_coeff[5]*b.temperature_red**2 +
                         b.entr_coeff[6]*b.pressure_red**2 +
                         b.entr_coeff[7]*b.temperature_red**3 +
                         b.entr_coeff[8]*b.temperature_red**2 *
                         b.pressure_red +
                         b.entr_coeff[9]*b.temperature_red *
                         b.pressure_red**2 +
                         b.entr_coeff[10]*b.pressure_red**3 +
                         b.entr_coeff[11]*log(b.temperature_red) +
                         b.entr_coeff[12]*b.temperature_red *
                         log(b.temperature_red) +
                         b.entr_coeff[13]*b.pressure_red *
                         log(b.temperature_red) +
                         b.entr_coeff[14]*b.temperature_red**2 *
                         log(b.temperature_red) +
                         b.entr_coeff[15]*b.pressure_red**2 *
                         log(b.temperature_red) +
                         b.entr_coeff[16]*b.temperature_red *
                         b.pressure_red*log(b.temperature_red) +
                         b.entr_coeff[17]*log(b.pressure_red) +
                         b.entr_coeff[18]*b.temperature_red *
                         log(b.pressure_red) +
                         b.entr_coeff[19]*b.pressure_red *
                         log(b.pressure_red) +
                         b.entr_coeff[20]*b.temperature_red**2 *
                         log(b.pressure_red) +
                         b.entr_coeff[21]*b.pressure_red**2 *
                         log(b.pressure_red) +
                         b.entr_coeff[22]*b.temperature_red *
                         b.pressure_red*log(b.pressure_red))
        try:
            # Try to build constraint
            self.entropy_correlation = Constraint(rule=entropy_correlation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.entr_mol)
            del self.entr_coeff
            self.del_component(self.entropy_correlation)
            raise

    def _heat_capacity_ratio(self):
        # Heat Capacity Ratio (from Spirax-Sarco via ALAMO)
        self.heat_capacity_ratio = Var(initialize=1.33,
                                       doc='Ratio of Heat Capacities')

        # Add reference to coefficients
        add_object_ref(self, "heat_capacity_ratio_coeff",
                       self.config.parameters.heat_capacity_ratio_coeff)

        def heat_capacity_ratio_correlation(b):
            return b.heat_capacity_ratio == \
                    (b.heat_capacity_ratio_coeff[1]*b.pressure*1e-6 +
                     b.heat_capacity_ratio_coeff[1]*b.temperature/100 +
                     b.heat_capacity_ratio_coeff[3]*log(b.pressure*1e-6) +
                     b.heat_capacity_ratio_coeff[4]*log(b.temperature/100) +
                     b.heat_capacity_ratio_coeff[5]*(b.temperature/100)**2 +
                     b.heat_capacity_ratio_coeff[6]*(b.temperature/100)**3 +
                     b.heat_capacity_ratio_coeff[7]*(b.pressure*1e-6)**0.5 +
                     b.heat_capacity_ratio_coeff[8]*(b.temperature/100)**0.5)
        try:
            # Try to build constraint
            self.heat_capacity_ratio_correlation = Constraint(
                                        rule=heat_capacity_ratio_correlation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.heat_capacity_ratio)
            del self.heat_capacity_ratio_coeff
            self.del_component(self.heat_capacity_ratio_correlation)
            raise

    def _dens_mol_phase(self):
        # Density from PV = ZRT
        self.dens_mol_phase = Var(self.phase_list, doc='Molar Density')

        def dens_mol_phase_calculation(b, p):
            return b.pressure == (b.dens_mol_phase[p]*b.compress_fact *
                                  b.gas_const*b.temperature)
        try:
            # Try to build constraint
            self.dens_mol_phase_calculation = Constraint(
                    self.phase_list,
                    rule=dens_mol_phase_calculation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.dens_mol_phase)
            self.del_component(self.dens_mol_phase_calculation)
            raise

    def _p_sat(self):
        # Vapour Pressure
        self.p_sat = Var(initialize=101325,
                         doc="Vapor pressure [Pa]")

        # Add reference to coefficients
        add_object_ref(self, "p_sat_coeff",
                       self.config.parameters.p_sat_coeff)

        def p_sat_correlation(b):
            return log(b.p_sat) == \
                            (b.p_sat_coeff[1] +
                             b.p_sat_coeff[2]/b.temperature +
                             b.p_sat_coeff[3]*b.temperature +
                             b.p_sat_coeff[4]*log(b.temperature) +
                             b.p_sat_coeff[5]*b.temperature**2)
        try:
            # Try to build constraint
            self.p_sat_correlation = Constraint(
                                            rule=p_sat_correlation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.p_sat)
            del self.p_sat_coeff
            self.del_component(self.p_sat_correlation)
            raise

    def _flow_mol(self):
        # Total Molar Flowrate
        self.flow_mol = Var(doc='Total molar flowrate')

        def flow_mol_calculation(b):
            return b.flow_mol == sum(b.flow_mol_comp[j]
                                     for j in b.component_list)
        try:
            # Try to build constraint
            self.flow_mol_calculation = Constraint(rule=flow_mol_calculation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.flow_mol)
            self.del_component(self.flow_mol_calculation)
            raise

    def _flow_vol(self):
        # Volumetric Flowrate
        self.flow_vol = Var(doc='Volumetric Flowrate')

        def volumetric_flow_calculation(b):
            return b.flow_vol*b.dens_mol_phase["Vap"] == \
                    sum(b.flow_mol_comp[j] for j in b.component_list)
        try:
            # Try to build constraint
            self.volumetric_flow_calculation = Constraint(
                                        rule=volumetric_flow_calculation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.flow_vol)
            self.del_component(self.volumetric_flow_calculation)
            raise

    def _mw(self):
        # Molecular Weight
        self.mw = Var(initialize=18.02, doc='Molecular weight')

        def mw_calculation(b):
            return b.mw == 18.02
        self.mw_calculation = Constraint(rule=mw_calculation)

    def _vap_frac(self):
        # Vapour Fraction
        self.vap_frac = Var(initialize=1.0, doc='Vapor Fraction')

        def vap_frac_calculation(b):
            return b.vap_frac == 1.0
        self.vap_frac_calculation = Constraint(rule=vap_frac_calculation)

    def _make_balance_terms(self):
        def material_terms(b, p, j):
            return b.flow_mol_comp[j]
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return b.flow_mol_comp["H2O"]*b.enth_mol
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return b.dens_mol_phase[p]
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return b.enth_mol*b.dens_mol_phase[p]
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"flow_mol_comp": b.flow_mol_comp,
                   "temperature": b.temperature,
                   "pressure": b.pressure}
        return members

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            logger.error('{} Temperature set below lower bound.'
                         .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            logger.error('{} Temperature set above upper bound.'
                         .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            logger.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            logger.error('{} Pressure set above upper bound.'.format(blk.name))
