##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration and test flowsheet for a dynamic flowsheet.

"""
from __future__ import division
from __future__ import print_function

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, SolverFactory

# Import IDAES core
from idaes.core import FlowsheetBlock, Stream

# Import Unit Model Modules
import tutorial_1_properties as rxn_properties
import tutorial_3_H2O_properties as H2O_properties

# Import Unit Model Modules
from idaes.models import Feed, Product, CSTR, HeatExchanger


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)

    # Add subflowsheets to flowsheet
    m.fs.rxn_properties = rxn_properties.PropertyParameterBlock()
    m.fs.H2O_properties = H2O_properties.PropertyParameterBlock()

    m.fs.config.default_property_package = m.fs.rxn_properties

    # Create unit models
    m.fs.Feed = Feed()
    m.fs.Tank1 = CSTR()
    m.fs.HX1 = HeatExchanger(side_2_property_package=m.fs.H2O_properties)

    m.fs.block = FlowsheetBlock(default_property_package=m.fs.H2O_properties)

    m.fs.block.Tank2 = CSTR(property_package=m.fs.rxn_properties)
    m.fs.block.HX2 = HeatExchanger(side_1_property_package=m.fs.rxn_properties)

    m.fs.Product = Product()

    # Post transform build
    m.fs.post_transform_build()

    # Make conenctors
    m.fs.stream_1 = Stream(source=m.fs.Feed.outlet,
                           destination=m.fs.Tank1.inlet)

    m.fs.stream_2 = Stream(source=m.fs.Tank1.outlet,
                           destination=m.fs.HX1.side_1_inlet)

    m.fs.stream_3 = Stream(source=m.fs.HX1.side_1_outlet,
                           destination=m.fs.block.Tank2.inlet)

    m.fs.block.stream_4 = Stream(source=m.fs.block.Tank2.outlet,
                                 destination=m.fs.block.HX2.side_1_inlet)

    m.fs.block.stream_5 = Stream(source=m.fs.block.HX2.side_1_outlet,
                                 destination=m.fs.Product.inlet)

    # Set inlet and operating conditions, and some initial conditions.
    m.fs.Feed.fix('flow_mol_comp', comp='a', value=1.0)
    m.fs.Feed.fix('flow_mol_comp', comp='b', value=2.0)
    m.fs.Feed.fix('flow_mol_comp', comp='c', value=0.1)
    m.fs.Feed.fix('flow_mol_comp', comp='d', value=0.0)
    m.fs.Feed.fix('flow_mol_comp', comp='e', value=0.0)
    m.fs.Feed.fix('flow_mol_comp', comp='f', value=0.0)
    m.fs.Feed.fix('temperature', value=303.15)
    m.fs.Feed.fix('pressure', value=101325)

    m.fs.Tank1.volume.fix(10.0)
    m.fs.Tank1.heat.fix(0.0)

    m.fs.HX1.side_2_inlet[:].vars["flow_mol"].fix(5.0)
    m.fs.HX1.side_2_inlet[:].vars["temperature"].fix(303.15)
    m.fs.HX1.side_2_inlet[:].vars["pressure"].fix(101325.0)

    m.fs.HX1.heat_transfer_coefficient.fix(100.0)
    m.fs.HX1.heat_transfer_area.fix(2.0)

    m.fs.block.Tank2.volume.fix(10.0)
    m.fs.block.Tank2.heat.fix(0.0)

    m.fs.block.HX2.side_2_inlet[:].vars["flow_mol"].fix(5.0)
    m.fs.block.HX2.side_2_inlet[:].vars["temperature"].fix(303.15)
    m.fs.block.HX2.side_2_inlet[:].vars["pressure"].fix(101325.0)

    m.fs.block.HX2.heat_transfer_coefficient.fix(100.0)
    m.fs.block.HX2.heat_transfer_area.fix(2.0)

    # Run model checks
#    m.fs.HX1.side_2_inlet[:].vars["temperature"].setlb(310.0)
#    m.fs.HX1.side_2.properties_in[0].display()
    m.fs.model_check()

    # Initialize Units
    m.fs.Feed.initialize()
    m.fs.Tank1.initialize(state_args={
            "flow_mol_comp": {
                "a": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["a"].value,
                "b": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["b"].value,
                "c": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["c"].value,
                "d": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["d"].value,
                "e": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["e"].value,
                "f": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["f"].value},
            "pressure": m.fs.Feed.outlet[0].vars["pressure"].value,
            "temperature": m.fs.Feed.outlet[0].vars["temperature"].value})
    m.fs.HX1.initialize(state_args_1={
            "flow_mol_comp": {
                "a": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["a"].value,
                "b": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["b"].value,
                "c": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["c"].value,
                "d": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["d"].value,
                "e": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["e"].value,
                "f": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["f"].value},
            "pressure": m.fs.Tank1.outlet[0].vars["pressure"].value,
            "temperature": m.fs.Tank1.outlet[0].vars["temperature"].value})

    m.fs.block.Tank2.initialize(state_args={
        "flow_mol_comp": {
            "a": m.fs.HX1.side_1_outlet[0].vars["flow_mol_comp"]["a"].value,
            "b": m.fs.HX1.side_1_outlet[0].vars["flow_mol_comp"]["b"].value,
            "c": m.fs.HX1.side_1_outlet[0].vars["flow_mol_comp"]["c"].value,
            "d": m.fs.HX1.side_1_outlet[0].vars["flow_mol_comp"]["d"].value,
            "e": m.fs.HX1.side_1_outlet[0].vars["flow_mol_comp"]["e"].value,
            "f": m.fs.HX1.side_1_outlet[0].vars["flow_mol_comp"]["f"].value},
        "pressure": m.fs.HX1.side_1_outlet[0].vars["pressure"].value,
        "temperature": m.fs.HX1.side_1_outlet[0].vars["temperature"].value})
    m.fs.block.HX2.initialize(state_args_1={
        "flow_mol_comp": {
            "a": m.fs.block.Tank2.outlet[0].vars["flow_mol_comp"]["a"].value,
            "b": m.fs.block.Tank2.outlet[0].vars["flow_mol_comp"]["b"].value,
            "c": m.fs.block.Tank2.outlet[0].vars["flow_mol_comp"]["c"].value,
            "d": m.fs.block.Tank2.outlet[0].vars["flow_mol_comp"]["d"].value,
            "e": m.fs.block.Tank2.outlet[0].vars["flow_mol_comp"]["e"].value,
            "f": m.fs.block.Tank2.outlet[0].vars["flow_mol_comp"]["f"].value},
        "pressure": m.fs.block.Tank2.outlet[0].vars["pressure"].value,
        "temperature": m.fs.block.Tank2.outlet[0].vars["temperature"].value})

    # Create a solver
    solver = SolverFactory('ipopt')
    results = solver.solve(m, tee=True)

    # Print results
    print(results)
    print()
    print("Results")
    m.fs.block.HX2.side_1_outlet.display()
    m.fs.block.HX2.side_2_outlet.display()

    # For testing purposes
    return(m, results)


if __name__ == "__main__":
    main()
