#!/usr/bin/env python
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Import property database from textfile(s).

See :meth:`PropertyData.import_csv`, for the expected format for data.
See :meth:`PropertyMetadata.import_text` for the expected format for metadata.

"""
import abc
import argparse
import csv
import json
import logging
import re
import sys

g_log = logging.getLogger(__name__)


class PropertyObject(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def as_json(self):
        """Return Python dict representation,
        that can be turned into JSON.
        """
        pass

    @staticmethod
    def get_file(file_or_path):
        if hasattr('read', file_or_path):
            return file_or_path
        return open(file_or_path)

class PropertyData(PropertyObject):
    """Class to import property data.

    """
    embedded_units = '(.*)\((.*)\)'
    STATE, PROPERTY = 'state', 'prop'

    def __init__(self,  state_columns):
        self._nstates = state_columns
        self._states = []
        self._properties = []
        self._initialized = False
        self._data = {}
        self._errors = {}
        self._types = {}
        self._units = {}
        self._error_type = {}
        self._col_names = {}

    def as_json(self):
        """Export the property data as JSON.

        Output will look like this::

            "T": {
                "type": "state",
                "units": "K",
                "values": [298.15, 298.15, ... , 353.15],
            },
            "Density Data": {
                "type": "property",
                "units": "g/cm^3",
                "values": [1.0053, 1.0188, .., ],
                "errors": [.00005, .., .00005],
                "error_type": "absolute"
            }

        Returns:
            (dict) JSON-able dict. Empty if data not initialized.
        """
        if not self._initialized:
            return {}
        result = {}
        for name, type_ in self._types.items():
            val = {
                'type': type_,
                'units': self._units[name],
                'values': self._data[name]
            }
            if self._errors[name] is not None:
                val['errors'] = self._errors[name]
                val['error_type'] = self._error_type[name]
            result[name] = val
        return result

    def import_csv(self, file_or_path):
        """Import the CSV data.

        Expected format of the  files is a header plus data rows.

        Header: Index-column,  Column-name(1), Error-column(1),  Column-name(2), Error-column(2), ..
        Data: <index>, <val>, <errval>, <val>, <errval>, ..

        Column-name is in the format "Name (units)"
        Error-column is in the format "Type Error", where "Type" is the error type.

        Args:
            file_or_path: Input file

        Returns:
            None
        """
        if self._initialized:
            raise RuntimeError('Attempt to import CSV into already '
                'initialized PropertyData instance')
        f = self.get_file(file_or_path)
        reader = csv.reader(f)
        ncol = -1
        for nrow, row in enumerate(reader):
            if nrow == 0:
                ncol = self._parse_header(row)
            else:
                self._parse_data(row, ncol)
        self._initialized = True

    def _parse_header(self, header):
        # header: assume <name>, <error>
        for i in range(1, len(header), 2):
            column = header[i]
            n = (i - 1) / 2 # column starting with 0
            name, units = self._parse_embedded_units(column)
            if n <= self._nstates:
                coltype = self.STATE
            else:
                coltype = self.PROPERTY
            self._types[name] = coltype
            self._data[name] = []
            self._errors[name] = []
            self._units[name] = units
            self._col_names[n] = name
            error_header = header[i+1]
            self._error_type[name] = error_header.strip().split()[0].lower()
        return (len(header) - 1) // 2

    def _parse_embedded_units(self, s):
        #print("Parse: {}".format(s))
        m = re.match(self.embedded_units, s)
        if m:
            name, units = m.groups()
        else:
            name, units = s, ''
        return name.strip(), units.strip()

    def _parse_data(self, row, num_columns):
        # XXX: replace missing value with string 'NaN'?!
        for i in range(1, num_columns * 2 + 1, 2):
            n = (i - 1) // 2
            name = self._col_names[n]
            rvalue, rerror = row[i], row[i+1]
            if rvalue == '':
                rvalueval = 'NaN'
            else:
                rvalueval = float(rvalue)
            self._data[name].append(rvalueval)
            if self._types[name] == self.PROPERTY:
                if rerror == '':
                    rerrorval = 'NaN'
                else:
                    rerrorval = float(rerror)
                self._errors[name].append(rerrorval)
            else:
                self._errors[name] = None


class PropertyMetadata(PropertyObject):
    """Class to import property metadata.
    """
    item_re = re.compile('\s*([a-zA-Z_-]*),"?(.*)"?\s*')

    def __init__(self, data_type=''):
        self._meta = {'datatype': data_type}

    def as_json(self):
        return self._meta

    def import_text(self, file_or_path):
        """Import metadata from simple text format.

        Example input::

            Source,Han, J., Jin, J., Eimer, D.A., Melaaen, M.C.,"Density of Water(1) + Monoethanolamine(2)
     + CO2(3) from (298.15 to 413.15) K and Surface Tension of Water(1) + Monethanolamine(2) from (
    303.15 to 333.15)K", J. Chem. Eng. Data, 2012, Vol. 57, pg. 1095-1103"
            Retrieval,"J. Morgan, date unknown"
            Notes,r is MEA weight fraction in aqueous soln. (CO2-free basis)

        Args:
            file_or_path: Input file

        Returns:
            None
        """
        f = self.get_file(file_or_path)
        for line in f:
            line = line.strip()
            if len(line) == 0 or line[0] == '#':
                continue
            m = self.item_re.match(line)
            if m is None:
                raise ValueError('Cannot parse metadata from: {}'.format(line))
            label, info = m.groups()
            label = label.lower()
            if label == 'source':
                self._parse_source(info)
            else:
                self._meta[label] = info

    def _parse_source(self, s):
        m = re.match('(.*),\s*"(.*)",\s*(.*)', s)
        if m is None:
            raise ValueError('Cannot parse source: {}'.format(s))
        auth, title, rest = m.groups()
        r = {'authors': auth, 'title': title, 'info': rest}
        self._meta.update(r)


def main():
    """Program entry point.
    """
    p = argparse.ArgumentParser(
        description='Import property database from text files')
    p.add_argument('file', help='Input CSV data file')
    p.add_argument('mfile', help='Input text metadata file')
    p.add_argument('-s', '--states', type=int, default=0,
                   help='Number of initial state columns')
    p.add_argument('-t', '--type', dest='dtype', default='MEA',
                   help='Property data type (default="%(default)s"')
    args = p.parse_args()
    #
    pd = PropertyData(args.states)
    pd.import_csv(args.file)
    #
    md = PropertyMetadata(data_type=args.dtype)
    md.import_text(args.mfile)
    #
    record = {
        'meta': md.as_json(),
        'data': pd.as_json()
    }
    print(json.dumps(record))

    return 0

if __name__ == '__main__':
    sys.exit(main())