##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Find and dump out values of properties stored in the
Google Cloud Datastore
"""
# System imports
import argparse
import json
import logging
import os
import sys
# Third-party imports
from google.cloud import datastore
# Local imports
from properties_upload import Datastore

_log = logging.getLogger('idaes.properties')
if not _log.handlers:
    _logh = logging.StreamHandler()
    _logh.setFormatter(logging.Formatter('%(asctime)s [%(levelname)s] %(message)s'))
    _log.addHandler(_logh)


def main():
    p = argparse.ArgumentParser(
        description='Dump properties from Google Cloud Datastore')
    p.add_argument('path', help='Path to properties, which should be '
                                'a dotted path with two components: '
                                '"<datatype>.<author_date>", e.g., '
                                '"MEA.Amundsen_2009".')
    p.add_argument('-k', '--key', default='key.json', metavar='FILE',
                   help='Key file with "service" key [%(default)s]')
    p.add_argument('-v', '--verbose', dest='vb', action='count', default=0,
                   help='Increase verbosity of output messages')
    args = p.parse_args()

    # Init logging
    if args.vb > 1:
        _log.setLevel(logging.DEBUG)
    elif args.vb > 0:
        _log.setLevel(logging.INFO)
    else:
        _log.setLevel(logging.WARN)

    # Check that we have a valid key file
    key_help = 'Use -k/--key option to ' \
               'indicate location of your Google Service Key, which should ' \
               'be a JSON file with {"type": "service_account"} as one of its ' \
               'key/value pairs.'
    if not os.path.isfile(args.key):
        p.error('Key file "{}" not found. {}'.format(args.key, key_help))
    # TODO: Some basic validation

    # Check that we have a valid path
    path_help = 'Path to properties should be ' \
                'a dotted path with two components: ' \
                '"<datatype>.<author_date>", e.g., ' \
                '"MEA.Amundsen_2009'
    f, r = args.path.find('.'), args.path.rfind('.')
    if f == -1:
        p.error('No dot separator found in path {p}: {h}'.format(p=args.path,
                                                                 h=path_help))
    if f != r:
        p.error('Too many dot separators in path {p}: {h}'.format(p=args.path,
                                                                  h=path_help))
    # Connect
    _log.info('Connect to Google Cloud Datastore')
    ds = Datastore(args.key)

    # Dump
    _log.info('Dump "Table" at: {}'.format(args.path))
    kind, key = args.path.split('.')
    key = key.lower()
    json_data = ds.fetch(kind, key)
    print(json.dumps(json_data))

    return 0

if __name__ == '__main__':
    sys.exit(main())