##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for dynamic tank example.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import SolverFactory, value
from pyomo.opt import SolverStatus, TerminationCondition
from idaes.ui.report import degrees_of_freedom

# Import file to be tested
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'dynamic_tank.py')
fe = import_file(example)


@pytest.fixture(scope="module")
def model_object():
    """Returns an instance of the model for testing.
    """
    m = fe.build_model()
    return m


# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


def test_build_model(model_object):
    # Try constructing the model and do a basic check
    assert value(model_object.fs_obj.Tank.area) == 1.0


def test_setInputs(model_object):
    m = model_object

    for t in m.fs_obj.time:
        # ---------------------------------------------------------------------
        # Unit Tank
        m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["a"].fix(1.0)
        m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["b"].fix(2.0)
        m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["c"].fix(0.1)
        m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["d"].fix(0.0)
        m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["e"].fix(0.0)
        m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["f"].fix(0.0)
        m.fs_obj.Tank.inlet[t].vars["enth_mol"].fix(-238.71)
        m.fs_obj.Tank.inlet[t].vars["pressure"].fix(101325.0)

        m.fs_obj.Tank.flow_coeff.fix(250000.0)
        m.fs_obj.Tank.area.fix(1.0)
        m.fs_obj.Tank.heat.fix(0.0)

        # Initial Condition - Steady-State
        m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "a"].fix(0.0)
        m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "b"].fix(0.0)
        m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "c"].fix(0.0)
        m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "d"].fix(0.0)
        m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "e"].fix(0.0)
        m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "f"].fix(0.0)
        m.fs_obj.Tank.holdup.energy_accumulation[0, "Liq"].fix(0.0)

        # Give an initial value for volume to help initialization
        m.fs_obj.Tank.volume[t] = 0.68

    # Check values were assigned properly
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["flow_mol_comp"]["a"]) ==
        1.0)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["flow_mol_comp"]["b"]) ==
        2.0)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["flow_mol_comp"]["c"]) ==
        0.1)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["flow_mol_comp"]["d"]) ==
        0.0)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["flow_mol_comp"]["e"]) ==
        0.0)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["flow_mol_comp"]["f"]) ==
        0.0)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["enth_mol"]) == -238.71)
    assert (
        value(model_object.fs_obj.Tank.inlet[0].vars["pressure"]) == 101325.0)

    assert value(model_object.fs_obj.Tank.flow_coeff[0]) == 250000.0
    assert value(model_object.fs_obj.Tank.area) == 1.0
    assert value(model_object.fs_obj.Tank.heat[0]) == 0.0

    assert value(model_object.fs_obj.Tank.volume[0]) == 0.68

    assert degrees_of_freedom(model_object) == 0


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization(model_object):
    results = fe.initialize_model(model_object, solver, stee=True)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_solution(model_object):
    m = model_object

    # Create a step change in feed
    for t in m.fs_obj.time:
        if t > 5.0:
            # After time = 5.0, 50% increase in feed rate
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["a"].fix(1.5)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["b"].fix(3.0)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["c"].fix(0.2)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["d"].fix(0.0)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["e"].fix(0.0)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["f"].fix(0.0)
            m.fs_obj.Tank.inlet[t].vars["enth_mol"].fix(-358.065)

    # Solve flowsheet
    results = fe.simulation1(model_object, solver, stee=True)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Check the solution numbers
    assert (pytest.approx(0.877, abs=1e-2) ==
            value(model_object.fs_obj.Tank.height[
                    model_object.fs_obj.time.last()]))
