##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for ports.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Expression, Set, Var
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        UnitBlockData, Holdup0D, \
                        PropertyParameterBase, PropertyBlockDataBase
from pyomo.common.config import ConfigValue


# -----------------------------------------------------------------------------
# Setup a basic unit model for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class UnitData(UnitBlockData):
    def build(self):
        pass


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    def build(self):
        super(PropertyParameterData, self).build()

    def get_supported_properties(self):
        return {'flow': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


@declare_process_block_class("PropertyBlock")
class PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        super(PropertyBlockData, self).build()
        self.phase_list = Set(initialize=["phase1", "phase2"])
        self.component_list = Set(initialize=["comp1", "comp2"])
        self.pressure = Var(initialize=1.0)
        self.temperature = Var(initialize=2.0)
        self.flow = Var(self.component_list, initialize=3.0)

        def material_terms(b, p, j):
            return 1
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return 2
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return 3
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return 4
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"a": b.pressure,
                   "b": b.temperature,
                   "c": b.flow}
        return members


@pytest.fixture()
def m():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=True)
    m.fs.prop = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.prop

    m.fs.u1 = Unit()
    m.fs.u1.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u1.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u1._setup_dynamics()
    m.fs.u1.holdup = Holdup0D()

    m.fs.u2 = Unit(dynamic=False)
    m.fs.u2.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u2.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u2._setup_dynamics()
    m.fs.u2.holdup = Holdup0D()
    return m


# -----------------------------------------------------------------------------
def test_fix_unfix_initial_conditions(m):
    m.fs.fix_initial_conditions()

    for p in m.fs.u1.holdup.phase_list:
        for j in m.fs.u1.holdup.component_list:
            assert m.fs.u1.holdup.material_accumulation[0, p, j].fixed
        assert m.fs.u1.holdup.energy_accumulation[0, p].fixed

    m.fs.unfix_initial_conditions()

    for p in m.fs.u1.holdup.phase_list:
        for j in m.fs.u1.holdup.component_list:
            assert not m.fs.u1.holdup.material_accumulation[0, p, j].fixed
        assert not m.fs.u1.holdup.energy_accumulation[0, p].fixed
