##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Test remote references feature of ProcessBlockData
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from unittest import TestCase
import os
import weakref
import gc

from pyomo.environ import ConcreteModel, Var, value, Expression
from idaes.core import ProcessBlockData, declare_process_block_class

__author__ = "John Eslick"


@declare_process_block_class("MyClass")
class MyClassData(ProcessBlockData):
    pass

class TestRemoteReferences(TestCase):
    def test_var_ref(self):
        """
        Reference sorks as expected for a scalar Var
        """
        m = ConcreteModel()
        m.x = Var()
        m.b = MyClass()
        m.b.remote.x = m.x
        m.x.fix(5)
        self.assertIsInstance(m.b.remote.x, weakref.ProxyTypes)
        self.assertEqual(m.b.remote.x.value, 5)
        self.assertEqual(value(m.b.remote.x), 5)
        assert(m.b.remote.x.fixed)
        m.b.expr = Expression(expr = m.b.remote.x + 3)
        self.assertEqual(value(m.b.expr), 8)

    def test_var_delete(self):
        """
        Make sure the right thing happens if the origianl var is deleted, using
        the reference in an expression will cause it to stick arround, same as
        original component.
        """
        m = ConcreteModel()
        m.x = Var()
        m.x.fix(5)
        m.b = MyClass()
        m.b.remote.x = m.x
        m.del_component("x")
        gc.collect() # Who knows when this would happen normally if I don't do
                     # this this test will beat the garbage collector and fail.
        with self.assertRaises(ReferenceError) as context:
            x = m.b.remote.x.value

    def test_indexed_var(self):
        m = ConcreteModel()
        m.x = Var([1,2])
        m.b = MyClass()
        m.b.remote.x = m.x[1]
        m.x[1].fix(5)
        self.assertIsInstance(m.b.remote.x, weakref.ProxyTypes)
        self.assertEqual(m.b.remote.x.value, 5)
        self.assertEqual(value(m.b.remote.x), 5)
        assert(m.b.remote.x.fixed)
        m.b.expr = Expression(expr = m.b.remote.x + 3)
        self.assertEqual(value(m.b.expr), 8)

    def test_not_referenceable(self):
        """
        Test that you get a type error if you cannot make a weakref to something
        """
        x = [1,2,3]
        b = MyClass()
        with self.assertRaises(TypeError) as context:
            b.remote.x = x
