##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for stream.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Expression, Set, value, Var
from pyomo.network import Port
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        UnitBlockData, Holdup0D, \
                        PropertyParameterBase, PropertyBlockDataBase
from idaes.core.stream import StreamData
from pyomo.common.config import ConfigValue


# -----------------------------------------------------------------------------
# Setup a basic unit model for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class UnitData(UnitBlockData):
    def build(self):
        pass


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    def build(self):
        super(PropertyParameterData, self).build()

    def get_supported_properties(self):
        return {'flow': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


@declare_process_block_class("PropertyBlock")
class PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        super(PropertyBlockData, self).build()
        self.phase_list = Set(initialize=["phase1", "phase2"])
        self.component_list = Set(initialize=["comp1", "comp2"])
        self.pressure = Var(initialize=1.0)
        self.temperature = Var(initialize=2.0)
        self.flow = Var(self.component_list, initialize=3.0)

        def material_terms(b, p, j):
            return 1
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return 2
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return 3
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return 4
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"a": b.pressure,
                   "b": b.temperature,
                   "c": b.flow}
        return members


@pytest.fixture()
def m():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.prop = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.prop

    m.fs.u1 = Unit()
    m.fs.u1.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u1.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u1._setup_dynamics()
    m.fs.u1.holdup = Holdup0D()
    m.fs.u1.build_outlets()

    m.fs.u2 = Unit()
    m.fs.u2.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u2.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u2._setup_dynamics()
    m.fs.u2.holdup = Holdup0D()
    m.fs.u2.build_inlets()
    return m
# -----------------------------------------------------------------------------


@declare_process_block_class("Stream")
class _Stream(StreamData):
    def build(self):
        pass


def test_config(m):
    m.fs.stream = Stream()

    # Test that Stream accepts ports as source and destination args
    m.fs.stream.config.source = m.fs.u1.outlet
    m.fs.stream.config.destination = m.fs.u1.outlet

    # Test that source and destination reject invalid values
    with pytest.raises(ValueError):
        m.fs.stream.config.source = "foo"  # invalid str
    with pytest.raises(ValueError):
        m.fs.stream.config.source = 1  # invalid int
    with pytest.raises(ValueError):
        m.fs.stream.config.source = 1.0  # invalid float
    with pytest.raises(ValueError):
        m.fs.stream.config.source = {"1": 1}  # invalid dict
    with pytest.raises(ValueError):
        m.fs.stream.config.source = [1]  # invalid list
    with pytest.raises(ValueError):
        m.fs.stream.config.source = m.fs.u1  # invalid non-port object

    with pytest.raises(ValueError):
        m.fs.stream.config.destination = "foo"  # invalid str
    with pytest.raises(ValueError):
        m.fs.stream.config.destination = 1  # invalid int
    with pytest.raises(ValueError):
        m.fs.stream.config.destination = 1.0  # invalid float
    with pytest.raises(ValueError):
        m.fs.stream.config.destination = {"1": 1}  # invalid dict
    with pytest.raises(ValueError):
        m.fs.stream.config.destination = [1]  # invalid list
    with pytest.raises(ValueError):
        m.fs.stream.config.destination = m.fs.u1  # invalid object

    # Test index arguments accept strs
    m.fs.stream.config.source_idx = "foo"
    m.fs.stream.config.destination_idx = "bar"
    # Most things can be cast to str - can't test for other types


def test_set_time_domain(m):
    m.fs.stream = Stream()
    m.fs.stream._set_time_domain()

    assert m.fs.stream.time == m.fs.time


def test_set_time_domain_no_time(m):
    m.stream = Stream()
    with pytest.raises(AttributeError):
        m.fs.stream._set_time_domain()


def test_check_ports_no_source(m):
    m.fs.stream = Stream(destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    with pytest.raises(ValueError):
        m.fs.stream._check_ports()


def test_check_ports_no_destination(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet)
    m.fs.stream._set_time_domain()
    with pytest.raises(ValueError):
        m.fs.stream._check_ports()


def test_check_ports_source_idx_not_indexed(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         source_idx="1",
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    with pytest.raises(KeyError):
        m.fs.stream._check_ports()


def test_check_ports_source_indexed_no_source_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.prop = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.prop

    m.fs.u1 = Unit()
    m.fs.u1.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u1.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u1._setup_dynamics()
    m.fs.u1.holdup = Holdup0D()
    m.fs.u1.build_outlets(num_outlets=2)

    m.fs.u2 = Unit()
    m.fs.u2.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u2.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u2._setup_dynamics()
    m.fs.u2.holdup = Holdup0D()
    m.fs.u2.build_inlets()

    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    with pytest.raises(KeyError):
        m.fs.stream._check_ports()


def test_check_ports_destination_idx_not_indexed(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet,
                         destination_idx="1",)
    m.fs.stream._set_time_domain()
    with pytest.raises(KeyError):
        m.fs.stream._check_ports()


def test_check_ports_destination_indexed_no_destination_idx():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.prop = PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.prop

    m.fs.u1 = Unit()
    m.fs.u1.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u1.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u1._setup_dynamics()
    m.fs.u1.holdup = Holdup0D()
    m.fs.u1.build_outlets()

    m.fs.u2 = Unit()
    m.fs.u2.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u2.config.declare("property_package_args", ConfigValue(default={}))
    m.fs.u2._setup_dynamics()
    m.fs.u2.holdup = Holdup0D()
    m.fs.u2.build_inlets(num_inlets=2)

    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    with pytest.raises(KeyError):
        m.fs.stream._check_ports()


def test_check_ports_diff_len():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    # Mock up two Ports of different lengths
    m.fs.c1 = Port(m.fs.time, noruleinit=True)
    m.fs.c2 = Port(m.fs.time, noruleinit=True)
    m.fs.v1 = Var(m.fs.time)
    m.fs.v2 = Var(m.fs.time)
    m.fs.v3 = Var(m.fs.time)
    m.fs.v4 = Var(m.fs.time)

    for t in m.fs.time:
        m.fs.c1[t].add(m.fs.v1, "a")
        m.fs.c1[t].add(m.fs.v2, "b")
        m.fs.c1[t].add(m.fs.v3, "c")
        m.fs.c1[t].add(m.fs.v4, "d")
        m.fs.c2[t].add(m.fs.v1, "a")
        m.fs.c2[t].add(m.fs.v2, "b")
        m.fs.c2[t].add(m.fs.v3, "c")

    m.fs.stream = Stream(source=m.fs.c1,
                         destination=m.fs.c2)
    m.fs.stream._set_time_domain()
    with pytest.raises(AttributeError):
        m.fs.stream._check_ports()


def test_check_ports_diff_members():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    # Mock up two Ports of different names
    m.fs.c1 = Port(m.fs.time, noruleinit=True)
    m.fs.c2 = Port(m.fs.time, noruleinit=True)
    m.fs.v1 = Var(m.fs.time)
    m.fs.v2 = Var(m.fs.time)
    m.fs.v3 = Var(m.fs.time)

    for t in m.fs.time:
        m.fs.c1[t].add(m.fs.v1, "d")
        m.fs.c1[t].add(m.fs.v2, "e")
        m.fs.c1[t].add(m.fs.v3, "f")
        m.fs.c2[t].add(m.fs.v1, "a")
        m.fs.c2[t].add(m.fs.v2, "b")
        m.fs.c2[t].add(m.fs.v3, "c")

    m.fs.stream = Stream(source=m.fs.c1,
                         destination=m.fs.c2)
    m.fs.stream._set_time_domain()
    with pytest.raises(AttributeError):
        m.fs.stream._check_ports()


def test_check_ports_diff_indexes():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    # Mock up two Ports with differnt indexes on members
    m.fs.c1 = Port(m.fs.time, noruleinit=True)
    m.fs.c2 = Port(m.fs.time, noruleinit=True)
    m.fs.v1 = Var(m.fs.time)
    m.fs.v2 = Var(m.fs.time)
    m.fs.v3 = Var(m.fs.time)
    m.fs.v4 = Var(m.fs.time, m.fs.time)

    for t in m.fs.time:
        m.fs.c1[t].add(m.fs.v1, "a")
        m.fs.c1[t].add(m.fs.v2, "b")
        m.fs.c1[t].add(m.fs.v3, "c")
        m.fs.c2[t].add(m.fs.v1, "a")
        m.fs.c2[t].add(m.fs.v2, "b")
        m.fs.c2[t].add(m.fs.v4, "c")

    m.fs.stream = Stream(source=m.fs.c1,
                         destination=m.fs.c2)
    m.fs.stream._set_time_domain()
    with pytest.raises(AttributeError):
        m.fs.stream._check_ports()


def test_build_constraints(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()

    assert hasattr(m.fs.stream, "a_equality")
    assert hasattr(m.fs.stream, "b_equality")
    assert hasattr(m.fs.stream, "c_equality")


def test_activate_deactivate(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()

    # Check Block is active
    assert m.fs.stream.active

    # Deactivate Block and assert
    m.fs.stream.deactivate()
    assert m.fs.stream.active is False

    # Reactivagte Block and assert
    m.fs.stream.activate()
    assert m.fs.stream.active


def test_activate_deactivate_constraint(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()

    # Check Block and Constraints are active
    assert m.fs.stream.active
    assert m.fs.stream.a_equality.active
    assert m.fs.stream.b_equality.active
    assert m.fs.stream.c_equality.active

    # Deactivate Constraints and assert
    m.fs.stream.deactivate("a")
    m.fs.stream.deactivate("c")
    assert m.fs.stream.active
    assert m.fs.stream.a_equality.active is False
    assert m.fs.stream.b_equality.active
    assert m.fs.stream.c_equality.active is False

    # Reactivagte Block and assert
    m.fs.stream.activate("c")
    assert m.fs.stream.active
    assert m.fs.stream.a_equality.active is False
    assert m.fs.stream.b_equality.active
    assert m.fs.stream.c_equality.active


def test_display(m):
    # Call display and to ensure no errors occur
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream.display()


def test_build_vars(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()

    assert hasattr(m.fs.stream, "a")
    assert hasattr(m.fs.stream, "b")
    assert hasattr(m.fs.stream, "c")


def test_build_vars_name_conflict(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream.a = 1
    with pytest.raises(AttributeError):
        m.fs.stream._build_vars()


def test_VarDict_fix1(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    m.fs.stream.a.fix(10)

    for t in m.fs.time:
        assert value(m.fs.stream.a[t][0]) == 1
        assert value(m.fs.stream.a[t][1]) == 10


def test_VarDict_fix2(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    m.fs.stream.a.fix(value=10, side='destination')

    for t in m.fs.time:
        assert value(m.fs.stream.a[t][0]) == 1
        assert value(m.fs.stream.a[t][1]) == 10


def test_VarDict_fix3(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    m.fs.stream.a.fix(value=10, side='source')

    for t in m.fs.time:
        assert value(m.fs.stream.a[t][0]) == 10
        assert value(m.fs.stream.a[t][1]) == 1


def test_VarDict_display1(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    m.fs.stream.a.display()


def test_VarDict_display2(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    m.fs.stream.a.display(side="source")


def test_VarDict_display3(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    m.fs.stream.a.display(side="destination")


def test_VarDict_display_wrong_side(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    with pytest.raises(ValueError):
        m.fs.stream.a.display(side="foo")


def test_converged_true(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()
    assert m.fs.stream.converged()


def test_converged_false(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()

    m.fs.u2.holdup.properties_in[0].pressure = 10
    assert m.fs.stream.converged() is False


def test_converged_tolerance(m):
    m.fs.stream = Stream(source=m.fs.u1.outlet,
                         destination=m.fs.u2.inlet)
    m.fs.stream._set_time_domain()
    m.fs.stream._check_ports()
    m.fs.stream._build_constraints()
    m.fs.stream._build_vars()

    m.fs.u2.holdup.properties_in[0].pressure = 1+1e-4
    assert m.fs.stream.converged() is False
    assert m.fs.stream.converged(tolerance=1e-4)
