##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests that tutorials run.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import SolverFactory
from pyomo.opt import SolverStatus, TerminationCondition
from idaes.ui.report import degrees_of_freedom

# Import file to be tested
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'tutorials')


# See if ipopt is available and set up solver
if not SolverFactory('ipopt').available():
    solver = None
else:
    solver = True


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_tutorial_1():
    f = import_file(join(example, "Tutorial_1_Basic_Flowsheets"))
    m, results = f.main()

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert degrees_of_freedom(m) == 0

    assert (m.fs.Product.product[0].vars["flow_mol_comp"]["a"].value ==
            pytest.approx(0.216, abs=1e-2))
    assert (m.fs.Product.product[0].vars["flow_mol_comp"]["b"].value ==
            pytest.approx(0.978, abs=1e-2))
    assert (m.fs.Product.product[0].vars["flow_mol_comp"]["c"].value ==
            pytest.approx(0.332, abs=1e-2))
    assert (m.fs.Product.product[0].vars["flow_mol_comp"]["d"].value ==
            pytest.approx(0.244, abs=1e-2))
    assert (m.fs.Product.product[0].vars["flow_mol_comp"]["e"].value ==
            pytest.approx(0.011, abs=1e-2))
    assert (m.fs.Product.product[0].vars["flow_mol_comp"]["f"].value ==
            pytest.approx(0.534, abs=1e-2))
    assert (m.fs.Product.product[0].vars["pressure"].value ==
            pytest.approx(101325, abs=1e-2))
    assert (m.fs.Product.product[0].vars["temperature"].value ==
            pytest.approx(402.146, abs=1e-2))


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_tutorial_2():
    f = import_file(join(example, "Tutorial_2_Basic_Flowsheet_Optimization"))
    m, results = f.main()

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert degrees_of_freedom(m) == 2

    assert (m.fs.Tank1.heat[0].value == pytest.approx(-7364.400, abs=1e-2))
    assert (m.fs.Tank2.heat[0].value == pytest.approx(-4201.702, abs=1e-2))


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_tutorial_3():
    f = import_file(join(example, "Tutorial_3_Advanced_Flowsheets"))
    m, results = f.main()

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert degrees_of_freedom(m) == 0

    assert (m.fs.block.HX2.side_1_outlet[0].vars["flow_mol_comp"]["a"].value ==
            pytest.approx(0.275, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["flow_mol_comp"]["b"].value ==
            pytest.approx(1.047, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["flow_mol_comp"]["c"].value ==
            pytest.approx(0.323, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["flow_mol_comp"]["d"].value ==
            pytest.approx(0.233, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["flow_mol_comp"]["e"].value ==
            pytest.approx(0.010, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["flow_mol_comp"]["f"].value ==
            pytest.approx(0.487, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["pressure"].value ==
            pytest.approx(101325, abs=1e-2))
    assert (m.fs.block.HX2.side_1_outlet[0].vars["temperature"].value ==
            pytest.approx(324.833, abs=1e-2))

    assert (m.fs.block.HX2.side_2_outlet[0].vars["flow_mol"].value ==
            pytest.approx(5.0, abs=1e-2))
    assert (m.fs.block.HX2.side_2_outlet[0].vars["pressure"].value ==
            pytest.approx(101325, abs=1e-2))
    assert (m.fs.block.HX2.side_2_outlet[0].vars["temperature"].value ==
            pytest.approx(318.794, abs=1e-2))


NO_SLOW_TESTS = False


@pytest.mark.skipif(NO_SLOW_TESTS or solver is None,
                    reason="Solver not available OR slow tests disabled")
def test_tutorial_4():
    f = import_file(join(example, "Tutorial_4_Dynamic_Flowsheets"))
    m, results = f.main()

    assert degrees_of_freedom(m) == 0

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok


@pytest.mark.skipif(NO_SLOW_TESTS or solver is None,
                    reason="Solver not available OR slow tests disabled")
def test_tutorial_5():
    f = import_file(join(example, "Tutorial_5_Modifying_Unit_Models"))
    m, results = f.main()

    assert degrees_of_freedom(m) == 0

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok
