##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""Utility functions for implementing outer approximation.

Jacobians are calculated using the Pyomo core differentiate function.

Implementation of OA formulation given in...
[citation needed, maybe the Viswanathan & Grossmann 1990 paper?]

In the current implementation, new constraint names for the cuts are
generated using the local_name attribute of the original nonlinear
constraint. This presents a problem because of potential name conflicts.

TODO: add support for Constraint of form a <= x <= b
TODO: add support for nonlinearities in the objective function

"""
from __future__ import division, print_function

from math import copysign

from pyomo.core.base import expr as EXPR
from pyomo.core.base.symbolic import differentiate
from pyomo.environ import (Constraint, NonNegativeReals, Objective, Set, Var,
                           maximize, minimize, value)

__author__ = "Qi Chen <qichen@andrew.cmu.edu>"


def apply_OA(nl_set, oa_block, jacs):
    """Apply outer approximation scheme.

    Uses augmented Lagrangian/equality relaxation scheme.

    Args:
        nl_set (set): Set of nonlinear constraints
        oa_block (Block): Pyomo block to hold generated OA cuts
        jacs (dict): dictionary in which to store jacobian information

    Returns:
        None

    Raises:
        ValueError: if a variable naming conflict exists

    TODO: possibility of naming conflicts due to use of local_name

    """
    # Set up outer approximation iterations set
    oa_iter = oa_block.oa_iter = Set()
    for con_data in nl_set:
        if not con_data.active:
            continue   # if constraint is inactive, skip it
        # compute the jacobian for this constraint
        # determine list of variables constained in the constraint
        constraint_vars = list(EXPR.identify_variables(con_data.body))
        # get a list of jacobians with respect to each constraint
        jac_list = differentiate(con_data.body, wrt_list=constraint_vars)
        # store the jacobian expressions in jacs. Note that zip takes the
        # equal-length lists constraint_vars and jac_list and creates a list of
        # tuples with one element from each list.
        jacs[con_data] = {
            id(var): jac for var, jac in zip(constraint_vars, jac_list)}

        # generate empty constraints for holding OA constraints
        constraint_name = con_data.parent_component().local_name
        if not hasattr(oa_block, constraint_name):
            if not con_data.parent_component().is_indexed():
                new_constr = Constraint(oa_iter)
            else:
                new_constr = Constraint(
                    oa_iter, con_data.parent_component()._index)
            setattr(oa_block, constraint_name, new_constr)
            # Generate slack variables for OA/ER
            new_slack_var = Var(oa_iter, domain=NonNegativeReals, initialize=0,
                                bounds=(0, 1000))
            setattr(oa_block, '_slack_' + constraint_name, new_slack_var)


def add_oa_constraints(iter_num, oa_block, jacs, exists_var, disagg_vars={}):
    """Add outer approximation constraints to the model.

    OA constraints are added based on the current model values and active
    constraints, as well as the duals returned from the NLP solver.

    """
    # Add new iteration to the set of iterations
    oa_block.oa_iter.add(iter_num)

    # Determine the objective sense value.
    # First, create generator over the active model objectives
    objs = oa_block.model().component_objects(
        ctype=Objective, active=True, descend_into=True)
    # fetch the sense of the first objective encountered
    obj_sense = next(objs).sense
    if next(objs, None) is not None:
        # If multiple active objectives exist, raise an error
        raise NotImplementedError(
            'Not sure how to handle dual values with multiple objectives.'
            ' Unable to create OA cuts.')
    # set the sign adjustment variable based on the objective sense
    if obj_sense == minimize:
        sign_adjust = -1
    elif obj_sense == maximize:
        sign_adjust = 1
    else:
        raise ValueError('Unexpected objective sense: ' + obj_sense)

    def _getvar(pyo_var):
        return disagg_vars.get(pyo_var.local_name, pyo_var)

    for con_data in jacs:
        con_name = con_data.parent_component().local_name  # constraint name
        oa_constr = getattr(oa_block, con_name)
        slack_var = getattr(oa_block, '_slack_' + con_name)
        # generate the appropriate index for the additional OA constraint
        if con_data.parent_component().is_indexed():
            con_data_index = con_data.index()
            if not isinstance(con_data_index, tuple):
                con_data_index = (con_data_index,)
            indx = (iter_num,) + con_data_index
        else:
            indx = iter_num
        dual = oa_block.model().dual[con_data]
        # verify that all variables have a lower bound, because we subtract by
        # it later
        var_list = list(EXPR.identify_variables(con_data.body))
        for var in var_list:
            if _getvar(var).lb is None:
                raise ValueError(
                    "{} needs to have a lower bound for proper OA cut"
                    " generation".format(_getvar(var).name))
        # Add the new OA constraint
        oa_constr.add(
            indx,
            (None,
             copysign(1, sign_adjust * dual) *
             sum(_get_jacobian(con_data, var, jacs) *
                 (_getvar(var) - _getvar(var).lb)
                 for var in var_list) -
             (copysign(1, sign_adjust * dual) * exists_var *
              sum(_get_jacobian(con_data, var, jacs) *
                  (value(var) - _getvar(var).lb)
                  for var in var_list) +
              slack_var[iter_num]),
             0))


def _get_jacobian(constr, var, jacs):
    """Retrieve the jacobian.

    Returns the derivative value for the constraint 'constr' with respect to
    variable 'var' with the current model values.

    """
    return value(jacs[constr][id(var)])
