##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration and test flowsheet for basic unit models.

"""
from __future__ import division
from __future__ import print_function

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, SolverFactory, \
            TransformationFactory, value

# Import IDAES core
from idaes.core import FlowsheetBlockData, declare_process_block_class

# Import Property Modules
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
prop_module = join(dirname(abspath(__file__)), '..', '..', '..', '..', 'examples',
                   'core', 'examples', 'steam_properties_demo.py')
properties = import_file(prop_module)

# Import Unit Model Modules
from idaes.models.pressure_changer import PressureChanger as Compressor

__author__ = "Andrew Lee"

# Set up logger
logger = logging.getLogger('idaes.flowsheet')


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    """
    Create the flowsheet class.  Contains all the unit models and
    connections between them.  Also contains the main Pyomo model and
    solver.
    """
    def __init__(self, *args, **kwargs):
        """
        Create a flowsheet model.
        """
        super(_Flowsheet, self).__init__(*args, **kwargs)

    def build(self):
        """
        Make the flowsheet object, fix some variables, and solve the problem
        """
        # Call UnitModel.build to setup dynamics
        super(_Flowsheet, self).build()

        # Add property packages to flowsheet library
        self.steam_props = properties.PropertyParameterBlock()

        # Create unit models
        self.C01 = Compressor(thermodynamic_assumption="isentropic",
                              include_holdup=False,
                              property_package=self.steam_props)

        # Call post_transform_build (needed for now)
        self.post_transform_build()


def build_model():
    """
    Construct flowsheet object
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Creat flowsheet object and attach to Concrete Model
    m.fs_obj = Flowsheet(dynamic=False)

    return m


def setInputs(m):
    """
    Set inlet and operating conditions.
    """

    for t in m.fs_obj.time:
        # ---------------------------------------------------------------------
        # Unit C01
        m.fs_obj.C01.inlet[t].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
        m.fs_obj.C01.inlet[t].vars["temperature"].fix(866.5)
        m.fs_obj.C01.inlet[t].vars["pressure"].fix(1e7)

        m.fs_obj.C01.deltaP.fix(1e6)
        m.fs_obj.C01.efficiency_isentropic.fix(0.83)


def initialize_model(m, solver, stee=False):
    """
    Set up and execute initializeation routine for model
    """
    # Initialize Units
    m.fs_obj.C01.initialize()

    # Solve flowsheet
    results = solver.solve(m.fs_obj, tee=stee)

    return results


def simulation1(m, solver, stee=False):
    # Solve flowsheet
    results = solver.solve(m, tee=stee)

    return results


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = build_model()

    # Fix variables
    setInputs(m)

    # Create a solver
    stee = True
    opt = SolverFactory('ipopt')
    opt.options = {'tol': 1e-6,
                   'mu_init': 1e-8,
                   'bound_push': 1e-8}

    m.fs_obj.model_check()

    # Solve flowsheet
    logger.info("Begin Flowsheet initialisation.")
    initialize_model(m, opt, stee)
    logger.info("Flowsheet initialisation finished.")

    # Run simulation
    results = simulation1(m, opt, stee)

    # Print results
    print(results)


if __name__ == "__main__":
    main()
