##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Feed unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        HoldupStatic
from idaes.models.feed import Feed, FeedData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'steam_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("FeedFrame")
class _Feed(FeedData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.feed = FeedFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.feed.config) == 16
    assert not m.fs.feed.config.dynamic
    assert not m.fs.feed.config.include_holdup
    assert m.fs.feed.config.material_balance_type == 'component_phase'
    assert m.fs.feed.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.feed.config.momentum_balance_type == 'pressure'
    assert not m.fs.feed.config.has_rate_reactions
    assert not m.fs.feed.config.has_equilibrium_reactions
    assert not m.fs.feed.config.has_phase_equilibrium
    assert not m.fs.feed.config.has_mass_transfer
    assert not m.fs.feed.config.has_heat_transfer
    assert not m.fs.feed.config.has_work_transfer
    assert not m.fs.feed.config.has_pressure_change
    assert m.fs.feed.config.property_package == m.fs.props
    assert m.fs.feed.config.property_package_args == {}
    assert m.fs.feed.config.outlet_list is None
    assert m.fs.feed.config.num_outlets is None


def test_super_build(m):
    super(FeedData, m.fs.feed).build()
    assert not m.fs.feed.config.dynamic


def test_adding_holdup(m):
    super(FeedData, m.fs.feed).build()
    m.fs.feed.holdup = HoldupStatic()

    assert not m.fs.feed.holdup.config.include_holdup
    assert m.fs.feed.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.feed.holdup.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.feed.holdup.config.momentum_balance_type == 'pressure'
    assert not m.fs.feed.holdup.config.has_rate_reactions
    assert not m.fs.feed.holdup.config.has_equilibrium_reactions
    assert not m.fs.feed.holdup.config.has_phase_equilibrium
    assert not m.fs.feed.holdup.config.has_mass_transfer
    assert not m.fs.feed.holdup.config.has_heat_transfer
    assert not m.fs.feed.holdup.config.has_work_transfer
    assert not m.fs.feed.holdup.config.has_pressure_change
    assert m.fs.feed.holdup.config.property_package == m.fs.props
    assert m.fs.feed.holdup.config.property_package_args == {}


def test_set_feed_references(m):
    super(FeedData, m.fs.feed).build()
    m.fs.feed.holdup = HoldupStatic()
    m.fs.feed._set_feed_references()

    assert hasattr(m.fs.feed, "feed")
    for k in m.fs.feed.feed[0].vars.keys():
        assert k in ["flow_mol_comp", "pressure", "temperature"]


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.feed = Feed(property_package=m2.fs.props)
    return m2


def test_post_transform_build_SO(m2):
    m2.fs.feed.post_transform_build()

    assert hasattr(m2.fs.feed, "outlet")

    # Test that outlet has correct indexing sets
    for k in m2.fs.feed.outlet:
        assert k in [0]


def test_post_transform_build_outlet_list(m2):
    m2.fs.feed.config.outlet_list = ["a", "b"]
    m2.fs.feed.post_transform_build()

    assert hasattr(m2.fs.feed, "outlet")

    # Test that outlet has correct indexing sets
    for k in m2.fs.feed.outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_outlet2(m2):
    m2.fs.feed.config.num_outlets = 2
    m2.fs.feed.post_transform_build()

    assert hasattr(m2.fs.feed, "outlet")

    # Test that outlet has correct indexing sets
    for k in m2.fs.feed.outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test user interface methods
def test_display():
    # Test that no errors occur when using display
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.feed = Feed(property_package=m.fs.props)
    m.fs.post_transform_build()
    m.fs.feed.display()


def test_fix_and_unfix_no_time():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.feed = Feed(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.feed.fix('flow_mol_comp', comp='H2O', value=27.5e3)
    assert m.fs.feed.holdup.properties[0].flow_mol_comp["H2O"].fixed
    assert m.fs.feed.holdup.properties[0].flow_mol_comp["H2O"].value == 27.5e3
    m.fs.feed.fix('temperature', value=866.5)
    assert m.fs.feed.holdup.properties[0].temperature.fixed
    assert m.fs.feed.holdup.properties[0].temperature.value == 866.5

    m.fs.feed.unfix('flow_mol_comp', comp='H2O')
    assert not m.fs.feed.holdup.properties[0].flow_mol_comp["H2O"].fixed
    m.fs.feed.unfix('temperature')
    assert not m.fs.feed.holdup.properties[0].temperature.fixed


def test_fix_and_unfix_set_time():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.feed = Feed(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.feed.fix('flow_mol_comp', comp='H2O', value=27.5e3, time=0)
    assert m.fs.feed.holdup.properties[0].flow_mol_comp["H2O"].fixed
    assert m.fs.feed.holdup.properties[0].flow_mol_comp["H2O"].value == 27.5e3
    m.fs.feed.fix('temperature', value=866.5, time=0)
    assert m.fs.feed.holdup.properties[0].temperature.fixed
    assert m.fs.feed.holdup.properties[0].temperature.value == 866.5

    m.fs.feed.unfix('flow_mol_comp', comp='H2O', time=0)
    assert not m.fs.feed.holdup.properties[0].flow_mol_comp["H2O"].fixed
    m.fs.feed.unfix('temperature', time=0)
    assert not m.fs.feed.holdup.properties[0].temperature.fixed


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.feed = Feed(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.feed.fix('flow_mol_comp', comp='H2O', value=27.5e3)
    m.fs.feed.fix('temperature', value=866.5)
    m.fs.feed.fix('pressure', value=2.97e7)

    assert degrees_of_freedom(m) == 0

    m.fs.feed.initialize()

    assert (pytest.approx(27500.0, abs=1e-2) ==
            m.fs.feed.outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(866.5, abs=1e-2) ==
            m.fs.feed.outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.feed.outlet[0].vars["pressure"].value)
