##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Flash unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Expression, Set, SolverFactory, \
                        Var, value
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D, PropertyParameterBase, PropertyBlockDataBase
from idaes.models.flash import Flash, FlashData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'phase_equilibrium_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("FlashFrame")
class _Flash(FlashData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.flash = FlashFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.flash.config) == 20
    assert m.fs.flash.config.dynamic == 'use_parent_value'
    assert not m.fs.flash.config.include_holdup
    assert m.fs.flash.config.material_balance_type == 'component_phase'
    assert m.fs.flash.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.flash.config.momentum_balance_type == 'pressure'
    assert not m.fs.flash.config.has_rate_reactions
    assert not m.fs.flash.config.has_equilibrium_reactions
    assert m.fs.flash.config.has_phase_equilibrium
    assert not m.fs.flash.config.has_mass_transfer
    assert m.fs.flash.config.has_heat_transfer
    assert not m.fs.flash.config.has_work_transfer
    assert m.fs.flash.config.has_pressure_change
    assert m.fs.flash.config.property_package == m.fs.props
    assert m.fs.flash.config.property_package_args == {}
    assert m.fs.flash.config.inlet_list is None
    assert m.fs.flash.config.num_inlets is None
    assert m.fs.flash.config.outlet_list == ["Vap", "Liq"]
    assert m.fs.flash.config.num_outlets is None
    assert m.fs.flash.config.outlet_type == "generic"
    assert m.fs.flash.config.energy_split_type == "temperature"


def test_super_build(m):
    super(FlashData, m.fs.flash).build()
    assert not m.fs.flash.config.dynamic


def test_adding_holdup(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()

    assert not m.fs.flash.holdup.config.include_holdup
    assert m.fs.flash.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.flash.holdup.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.flash.holdup.config.momentum_balance_type == 'pressure'
    assert not m.fs.flash.holdup.config.has_rate_reactions
    assert not m.fs.flash.holdup.config.has_equilibrium_reactions
    assert m.fs.flash.holdup.config.has_phase_equilibrium
    assert not m.fs.flash.holdup.config.has_mass_transfer
    assert m.fs.flash.holdup.config.has_heat_transfer
    assert not m.fs.flash.holdup.config.has_work_transfer
    assert m.fs.flash.holdup.config.has_pressure_change
    assert m.fs.flash.holdup.config.property_package == m.fs.props
    assert m.fs.flash.holdup.config.property_package_args == {}


def test_set_geometry_include_holdup_false(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()

    assert not hasattr(m.fs.flash, "volume")


def test_set_geometry_include_holdup_true(m):
    m.fs.flash.config.include_holdup = True
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()

    assert hasattr(m.fs.flash, "volume")
    assert hasattr(m.fs.flash.holdup, "volume")
    assert m.fs.flash.volume == m.fs.flash.holdup.volume


def test_make_performance_has_heat_transfer_true(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             material_split_type="phase")
    m.fs.flash._make_performance()

    assert hasattr(m.fs.flash, "heat")


def test_make_performance_has_heat_transfer_false(m):
    m.fs.flash.config.has_heat_transfer = False
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             material_split_type="phase")
    m.fs.flash._make_performance()

    assert not hasattr(m.fs.flash, "heat")


def test_make_performance_has_pressure_change_true(m):
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             material_split_type="phase")
    m.fs.flash._make_performance()

    assert hasattr(m.fs.flash, "deltaP")


def test_make_performance_has_pressure_change_false(m):
    m.fs.flash.config.has_pressure_change = False
    super(FlashData, m.fs.flash).build()
    m.fs.flash.holdup = Holdup0D()
    m.fs.flash._set_geometry()
    m.fs.flash.build_outlets(outlets=m.fs.flash.config.outlet_list,
                             num_outlets=m.fs.flash.config.num_outlets,
                             material_split_type="phase")
    m.fs.flash._make_performance()

    assert not hasattr(m.fs.flash, "deltaP")


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.flash = Flash(property_package=m2.fs.props)
    return m2


def test_post_transform_build_no_args(m2):
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [0]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "Vap"), (0, "Liq")]

    # Check that split fractions have been set correctly
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Vap",
                                                             "Vap"].fixed
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Vap",
                                                             "Vap"].value == 1
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Liq",
                                                             "Liq"].fixed
    assert m2.fs.flash.holdup.outlet_splitter.split_fraction[0,
                                                             "Liq",
                                                             "Liq"].value == 1


def test_post_transform_build_inlet_list(m2):
    m2.fs.flash.config.inlet_list = ["a", "b"]
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "Vap"), (0, "Liq")]


def test_post_transform_build_num_inlets(m2):
    m2.fs.flash.config.num_inlets = 2
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "Vap"), (0, "Liq")]


def test_post_transform_build_outlet_list(m2):
    m2.fs.flash.config.outlet_list = ["a", "b"]
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [0]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "a"), (0, "b")]

    # Check that split fractions have not been fixed
    for k in m2.fs.flash.holdup.outlet_splitter.split_fraction[0, :, :]:
        assert not k.fixed


def test_post_transform_build_num_outlet2(m2):
    m2.fs.flash.config.outlet_list = None
    m2.fs.flash.config.num_outlets = 2
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash, "inlet")
    assert hasattr(m2.fs.flash, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.flash.inlet:
        assert k in [0]
    for k in m2.fs.flash.outlet:
        assert k in [(0, "1"), (0, "2")]

    # Check that split fractions have not been fixed
    for k in m2.fs.flash.holdup.outlet_splitter.split_fraction[0, :, :]:
        assert not k.fixed


def test_post_transform_build_direct_outlet(m2):
    m2.fs.flash.config.outlet_type = "direct"
    m2.fs.flash.post_transform_build()

    assert hasattr(m2.fs.flash.outlet[0, "Liq"], "pressure")
    assert hasattr(m2.fs.flash.outlet[0, "Liq"], "temperature")

    assert hasattr(m2.fs.flash.outlet[0, "Vap"], "pressure")
    assert hasattr(m2.fs.flash.outlet[0, "Vap"], "temperature")


# -----------------------------------------------------------------------------
# Setup test classes for testing direct outlet splitting
@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    def build(self):
        super(PropertyParameterData, self).build()
        self._test_flag = 1

    def get_supported_properties(self):
        if self._test_flag == 1:
            return {'flow': {'method': None, 'units': 'mol/s'},
                    'pressure': {'method': None, 'units': 'Pa'},
                    'temperature': {'method': None, 'units': 'K'},
                    'enth_mol': {'method': None, 'units': 'J/mol'},
                    'enth_mass': {'method': None, 'units': 'J/g'}}
        if self._test_flag == 3:
            return {'flow_mol_phase': {'method': None, 'units': 'mol/s'},
                    'mole_frac_phase': {'method': None, 'units': 'mol/s'},
                    'pressure': {'method': None, 'units': 'Pa'}}
        else:
            return {'flow_mol': {'method': None, 'units': 'mol/s'},
                    'pressure': {'method': None, 'units': 'Pa'},
                    'temperature': {'method': None, 'units': 'K'},
                    'enth_mol': {'method': None, 'units': 'J/mol'},
                    'enth_mass': {'method': None, 'units': 'J/g'},
                    'enth_mol_phase': {'method': None, 'units': 'J/mol'},
                    'enth_mass_phase': {'method': None, 'units': 'J/g'},
                    'flow_mol_phase': {'method': None, 'units': 'mol/s'},
                    'flow_mol_phase_comp': {'method': None, 'units': 'mol/s'},
                    'flow_mol_comp': {'method': None, 'units': 'mol/s'},
                    'flow_mass_phase_comp': {'method': None, 'units': 'g/s'},
                    'flow_mass_phase': {'method': None, 'units': 'g/s'},
                    'flow_mass_comp': {'method': None, 'units': 'g/s'},
                    'mole_frac': {'method': None, 'units': '-'},
                    'mass_frac': {'method': None, 'units': '-'},
                    'mole_frac_phase': {'method': None, 'units': '-'},
                    'mass_frac_phase': {'method': None, 'units': '-'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


@declare_process_block_class("PropertyBlock")
class PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        super(PropertyBlockData, self).build()
        self.phase_list = Set(initialize=["Vap", "Liq"])
        self.component_list = Set(initialize=["comp1", "comp2"])
        self.pressure = Var(initialize=1.0)
        self.temperature = Var(initialize=2.0)
        self.flow = Var(self.component_list, initialize=3.0)

        self._test_flag = 1

        self.enth_mol = Var(initialize=0.707)
        self.enth_mol_phase = Var(self.phase_list, initialize=0.707)
        self.enth_mass = Var(initialize=0.707)
        self.enth_mass_phase = Var(self.phase_list, initialize=0.707)
        self.flow_mol_phase = Var(self.phase_list, initialize=7)
        self.flow_mol_phase_comp = Var(self.phase_list,
                                       self.component_list,
                                       initialize=7)
        self.flow_mass = Var(initialize=7)
        self.flow_mass_comp = Var(self.component_list,
                                  initialize=7)
        self.flow_mass_phase = Var(self.phase_list, initialize=7)
        self.flow_mass_phase_comp = Var(self.phase_list,
                                        self.component_list,
                                        initialize=7)
        self.mole_frac_phase = Var(self.phase_list,
                                   self.component_list,
                                   initialize=0.5)
        self.mass_frac_phase = Var(self.phase_list,
                                   self.component_list,
                                   initialize=0.5)
        self.mass_frac = Var(self.component_list, initialize=0.5)

        def material_terms(b, p, j):
            return 1
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return 2
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return 3
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return 4
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        if b._test_flag == 1:
            members = {"a": b.pressure,
                       "b": b.temperature,
                       "c": b.flow}
        elif b._test_flag == 2:
            members = {"pressure": b.pressure,
                       "temperature": b.temperature,
                       "flow_mol": b.flow,
                       "mole_frac": 1}
        elif b._test_flag == 3:
            members = {"pressure": b.pressure,
                       "temperature": b.temperature,
                       "flow_mol_comp": b.flow}
        elif b._test_flag == 4:
            members = {"pressure": b.pressure,
                       "temperature": b.temperature,
                       "flow_mol_phase_comp": b.flow}
        elif b._test_flag == 5:
            members = {"pressure": b.pressure,
                       "temperature": b.temperature,
                       "flow_mol_phase": b.flow,
                       "mole_frac_phase": b.mole_frac_phase}
        elif b._test_flag == 10:
            members = {"pressure": b.pressure,
                       "enth_mol": b.enth_mol,
                       "flow_mol_phase": b.flow_mol_phase,
                       "mole_frac_phase": b.mole_frac_phase}
        elif b._test_flag == 11:
            members = {"pressure": b.pressure,
                       "enth_mol_phase": b.enth_mol_phase,
                       "flow_mol_phase_comp": b.flow_mol_phase_comp}
        elif b._test_flag == 12:
            members = {"pressure": b.pressure,
                       "enth_mass": b.enth_mass,
                       "flow_mass_comp": b.flow_mass_comp}
        elif b._test_flag == 13:
            members = {"pressure": b.pressure,
                       "enth_mass_phase": b.enth_mass_phase,
                       "flow_mass": b.flow_mass,
                       "mass_frac": b.mass_frac}
        else:
            members = {"pressure": b.pressure,
                       "temperature": b.temperature,
                       "flow_mol": b.flow}
        return members


@pytest.fixture()
def m3():
    """
    Build a simple flowsheet model for testing direct outlet splitting.
    """
    m3 = ConcreteModel()
    m3.fs = Flowsheet(dynamic=False)
    m3.fs.props = PropertyParameterBlock()
    return m3


def test_direct_splitting_invalid_states(m3):
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_1(m3):
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 2
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_2(m3):
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 3
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_3(m3):
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 4
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_4(m3):
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 5
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_5(m3):
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 6
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_6(m3):
    m3.fs.props._test_flag = 3
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 5
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_unsupported_props_7(m3):
    m3.fs.props._test_flag = 3
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 10
    m3.fs.flash.post_transform_build()
    assert hasattr(m3.fs.flash.holdup, "outlet_splitter")


def test_direct_splitting_non_default_outlets(m3):
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.flash = Flash(property_package=m.fs.props,
                       outlet_type="direct",
                       outlet_list=['1', '2', '3'],
                       num_outlets=4)
    m.fs.post_transform_build()

    assert m.fs.flash.config.outlet_list == ["Vap", "Liq"]
    assert m.fs.flash.config.num_outlets is None


def test_direct_splitting_1(m3):
    m3.fs.props._test_flag = 2
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 10
    m3.fs.flash.post_transform_build()

    assert not hasattr(m3.fs.flash.holdup, "outlet_splitter")

    assert (m3.fs.flash.outlet[0, 'Vap'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Vap'].enth_mol ==
            m3.fs.flash.holdup.properties_out[0].enth_mol_phase['Vap'])
    assert (m3.fs.flash.outlet[0, 'Vap'].flow_mol_phase['Vap'] ==
            m3.fs.flash.holdup.properties_out[0].flow_mol_phase['Vap'])
    assert (m3.fs.flash.outlet[0, 'Vap'].flow_mol_phase['Liq'] == 0)
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (m3.fs.flash.outlet[0, 'Vap'].mole_frac_phase['Vap', j] ==
                m3.fs.flash.holdup.properties_out[0].mole_frac_phase['Vap', j])
        assert (m3.fs.flash.outlet[0, 'Vap'].mole_frac_phase['Liq', j] == 0)

    assert (m3.fs.flash.outlet[0, 'Liq'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Liq'].enth_mol ==
            m3.fs.flash.holdup.properties_out[0].enth_mol_phase['Liq'])
    assert (m3.fs.flash.outlet[0, 'Liq'].flow_mol_phase['Liq'] ==
            m3.fs.flash.holdup.properties_out[0].flow_mol_phase['Liq'])
    assert (m3.fs.flash.outlet[0, 'Liq'].flow_mol_phase['Vap'] == 0)
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (m3.fs.flash.outlet[0, 'Liq'].mole_frac_phase['Liq', j] ==
                m3.fs.flash.holdup.properties_out[0].mole_frac_phase['Liq', j])
        assert (m3.fs.flash.outlet[0, 'Liq'].mole_frac_phase['Vap', j] == 0)


def test_direct_splitting_2(m3):
    m3.fs.props._test_flag = 2
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 11
    m3.fs.flash.post_transform_build()

    assert not hasattr(m3.fs.flash.holdup, "outlet_splitter")

    assert (m3.fs.flash.outlet[0, 'Vap'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Vap'].enth_mol_phase ==
            m3.fs.flash.holdup.properties_out[0].enth_mol_phase)
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (
            m3.fs.flash.outlet[0, 'Vap'].flow_mol_phase_comp['Vap', j] ==
            m3.fs.flash.holdup.properties_out[0].flow_mol_phase_comp['Vap', j])
        assert (m3.fs.flash.outlet[0, 'Vap'].flow_mol_phase_comp['Liq', j] ==
                0)

    assert (m3.fs.flash.outlet[0, 'Liq'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Liq'].enth_mol_phase ==
            m3.fs.flash.holdup.properties_out[0].enth_mol_phase)
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (
            m3.fs.flash.outlet[0, 'Liq'].flow_mol_phase_comp['Liq', j] ==
            m3.fs.flash.holdup.properties_out[0].flow_mol_phase_comp['Liq', j])
        assert (m3.fs.flash.outlet[0, 'Liq'].flow_mol_phase_comp['Vap', j] ==
                0)


def test_direct_splitting_3(m3):
    m3.fs.props._test_flag = 2
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 12
    m3.fs.flash.post_transform_build()

    assert not hasattr(m3.fs.flash.holdup, "outlet_splitter")

    assert (m3.fs.flash.outlet[0, 'Vap'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Vap'].enth_mass ==
            m3.fs.flash.holdup.properties_out[0].enth_mass_phase['Vap'])
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (
            m3.fs.flash.outlet[0, 'Vap'].flow_mass_comp[j] ==
            m3.fs.flash.holdup.properties_out[0]
            .flow_mass_phase_comp['Vap', j])

    assert (m3.fs.flash.outlet[0, 'Liq'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Liq'].enth_mass ==
            m3.fs.flash.holdup.properties_out[0].enth_mass_phase['Liq'])
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (
            m3.fs.flash.outlet[0, 'Liq'].flow_mass_comp[j] ==
            m3.fs.flash.holdup.properties_out[0]
            .flow_mass_phase_comp['Liq', j])


def test_direct_splitting_4(m3):
    m3.fs.props._test_flag = 2
    m3.fs.flash = Flash(property_package=m3.fs.props,
                        outlet_type="direct")
    m3.fs.flash.holdup.properties_out[0]._test_flag = 13
    m3.fs.flash.post_transform_build()

    assert not hasattr(m3.fs.flash.holdup, "outlet_splitter")

    assert (m3.fs.flash.outlet[0, 'Vap'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Vap'].enth_mass_phase ==
            m3.fs.flash.holdup.properties_out[0].enth_mass_phase)
    assert (m3.fs.flash.outlet[0, 'Vap'].flow_mass ==
            m3.fs.flash.holdup.properties_out[0].flow_mass_phase['Vap'])
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (m3.fs.flash.outlet[0, 'Vap'].mass_frac[j] ==
                m3.fs.flash.holdup.properties_out[0].mass_frac_phase['Vap', j])

    assert (m3.fs.flash.outlet[0, 'Liq'].pressure ==
            m3.fs.flash.holdup.properties_out[0].pressure)
    assert (m3.fs.flash.outlet[0, 'Liq'].enth_mass_phase ==
            m3.fs.flash.holdup.properties_out[0].enth_mass_phase)
    assert (m3.fs.flash.outlet[0, 'Liq'].flow_mass ==
            m3.fs.flash.holdup.properties_out[0].flow_mass_phase['Liq'])
    for j in m3.fs.flash.holdup.properties_out[0].component_list:
        assert (m3.fs.flash.outlet[0, 'Liq'].mass_frac[j] ==
                m3.fs.flash.holdup.properties_out[0].mass_frac_phase['Liq', j])


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.flash = Flash(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Vap", "CO2"].fix(0.1)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Vap", "H2O"].fix(0.1)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Vap", "N2"].fix(0.8)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Liq", "CO2"].fix(0.0)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Liq", "H2O"].fix(2.0)
    m.fs.flash.inlet[:].vars["temperature"].fix(300.0)
    m.fs.flash.inlet[:].vars["pressure"].fix(101325.0)

    m.fs.flash.heat.fix(0.0)
    m.fs.flash.deltaP.fix(0.0)

    assert degrees_of_freedom(m) == 0

    m.fs.flash.initialize()

    assert (
        pytest.approx(0.000328, abs=1e-5) ==
        m.fs.flash.outlet[0, "Liq"]
        .vars["flow_mol_phase_comp"]["Liq", "CO2"].value)
    assert (
        pytest.approx(2.068, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"]
        .vars["flow_mol_phase_comp"]["Liq", "H2O"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"]
        .vars["flow_mol_phase_comp"]["Vap", "CO2"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"]
        .vars["flow_mol_phase_comp"]["Vap", "H2O"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Liq"]
        .vars["flow_mol_phase_comp"]["Vap", "N2"].value)
    assert (pytest.approx(300.056, abs=1e-2) ==
            m.fs.flash.outlet[0, "Liq"].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.flash.outlet[0, "Liq"].vars["pressure"].value)

    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"]
        .vars["flow_mol_phase_comp"]["Liq", "CO2"].value)
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"]
        .vars["flow_mol_phase_comp"]["Liq", "H2O"].value)
    assert (
        pytest.approx(0.100, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"]
        .vars["flow_mol_phase_comp"]["Vap", "CO2"].value)
    assert (
        pytest.approx(0.0323, abs=1e-3) ==
        m.fs.flash.outlet[0, "Vap"]
        .vars["flow_mol_phase_comp"]["Vap", "H2O"].value)
    assert (
        pytest.approx(0.8, abs=1e-2) ==
        m.fs.flash.outlet[0, "Vap"]
        .vars["flow_mol_phase_comp"]["Vap", "N2"].value)
    assert (pytest.approx(300.056, abs=1e-2) ==
            m.fs.flash.outlet[0, "Vap"].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.flash.outlet[0, "Vap"].vars["pressure"].value)


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization_direct_outlet():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.flash = Flash(property_package=m.fs.props,
                       outlet_type="direct")
    m.fs.post_transform_build()

    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Vap", "CO2"].fix(0.1)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Vap", "H2O"].fix(0.1)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Vap", "N2"].fix(0.8)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Liq", "CO2"].fix(0.0)
    m.fs.flash.inlet[:].vars["flow_mol_phase_comp"]["Liq", "H2O"].fix(2.0)
    m.fs.flash.inlet[:].vars["temperature"].fix(300.0)
    m.fs.flash.inlet[:].vars["pressure"].fix(101325.0)

    m.fs.flash.heat.fix(0.0)
    m.fs.flash.deltaP.fix(0.0)

    assert degrees_of_freedom(m) == 0

    m.fs.flash.initialize()

    assert (
        pytest.approx(0.000328, abs=1e-5) ==
        value(m.fs.flash.outlet[0, "Liq"]
              .vars["flow_mol_phase_comp"]["Liq", "CO2"]))
    assert (
        pytest.approx(2.068, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Liq"]
              .vars["flow_mol_phase_comp"]["Liq", "H2O"]))
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Liq"]
              .vars["flow_mol_phase_comp"]["Vap", "CO2"]))
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Liq"]
              .vars["flow_mol_phase_comp"]["Vap", "H2O"]))
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Liq"]
              .vars["flow_mol_phase_comp"]["Vap", "N2"]))
    assert (pytest.approx(300.056, abs=1e-2) ==
            m.fs.flash.outlet[0, "Liq"].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.flash.outlet[0, "Liq"].vars["pressure"].value)

    assert (
        pytest.approx(0.000, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Vap"]
              .vars["flow_mol_phase_comp"]["Liq", "CO2"]))
    assert (
        pytest.approx(0.000, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Vap"]
              .vars["flow_mol_phase_comp"]["Liq", "H2O"]))
    assert (
        pytest.approx(0.100, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Vap"]
              .vars["flow_mol_phase_comp"]["Vap", "CO2"]))
    assert (
        pytest.approx(0.0323, abs=1e-3) ==
        value(m.fs.flash.outlet[0, "Vap"]
              .vars["flow_mol_phase_comp"]["Vap", "H2O"]))
    assert (
        pytest.approx(0.8, abs=1e-2) ==
        value(m.fs.flash.outlet[0, "Vap"]
              .vars["flow_mol_phase_comp"]["Vap", "N2"]))
    assert (pytest.approx(300.056, abs=1e-2) ==
            m.fs.flash.outlet[0, "Vap"].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.flash.outlet[0, "Vap"].vars["pressure"].value)
