##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Gibbs Reactor unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D
from idaes.models.gibbs_reactor import GibbsReactor, GibbsReactorData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'gibbs_reactor_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("GibbsReactorFrame")
class _GibbsReactor(GibbsReactorData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.gibbs = GibbsReactorFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.gibbs.config) == 18
    assert not m.fs.gibbs.config.dynamic
    assert not m.fs.gibbs.config.include_holdup
    assert m.fs.gibbs.config.material_balance_type == 'element_total'
    assert m.fs.gibbs.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.gibbs.config.momentum_balance_type == 'pressure'
    assert not m.fs.gibbs.config.has_rate_reactions
    assert m.fs.gibbs.config.has_equilibrium_reactions
    assert m.fs.gibbs.config.has_phase_equilibrium
    assert not m.fs.gibbs.config.has_mass_transfer
    assert m.fs.gibbs.config.has_heat_transfer
    assert not m.fs.gibbs.config.has_work_transfer
    assert not m.fs.gibbs.config.has_pressure_change
    assert m.fs.gibbs.config.property_package == m.fs.props
    assert m.fs.gibbs.config.property_package_args == {}
    assert m.fs.gibbs.config.inlet_list is None
    assert m.fs.gibbs.config.num_inlets is None
    assert m.fs.gibbs.config.outlet_list is None
    assert m.fs.gibbs.config.num_outlets is None


def test_super_build(m):
    super(GibbsReactorData, m.fs.gibbs).build()
    assert not m.fs.gibbs.config.dynamic


def test_adding_holdup(m):
    super(GibbsReactorData, m.fs.gibbs).build()
    m.fs.gibbs.holdup = Holdup0D()

    assert not m.fs.gibbs.holdup.config.include_holdup
    assert m.fs.gibbs.holdup.config.material_balance_type == 'element_total'
    assert m.fs.gibbs.holdup.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.gibbs.holdup.config.momentum_balance_type == 'pressure'
    assert not m.fs.gibbs.holdup.config.has_rate_reactions
    # Property package does not include equilibrium terms of either type
    assert not m.fs.gibbs.holdup.config.has_equilibrium_reactions
    assert not m.fs.gibbs.holdup.config.has_phase_equilibrium
    assert not m.fs.gibbs.holdup.config.has_mass_transfer
    assert m.fs.gibbs.holdup.config.has_heat_transfer
    assert not m.fs.gibbs.holdup.config.has_work_transfer
    assert not m.fs.gibbs.holdup.config.has_pressure_change
    assert m.fs.gibbs.holdup.config.property_package == m.fs.props
    assert m.fs.gibbs.holdup.config.property_package_args == {}


def test_set_geometry_include_holdup_false(m):
    super(GibbsReactorData, m.fs.gibbs).build()
    m.fs.gibbs.holdup = Holdup0D()
    m.fs.gibbs._set_geometry()

    assert not hasattr(m.fs.gibbs, "volume")


def test_set_geometry_include_holdup_true(m):
    m.fs.gibbs.config.include_holdup = True
    super(GibbsReactorData, m.fs.gibbs).build()
    m.fs.gibbs.holdup = Holdup0D()
    m.fs.gibbs._set_geometry()

    assert hasattr(m.fs.gibbs, "volume")
    assert hasattr(m.fs.gibbs.holdup, "volume")
    assert m.fs.gibbs.volume == m.fs.gibbs.holdup.volume


def test_make_performance(m):
    super(GibbsReactorData, m.fs.gibbs).build()
    m.fs.gibbs.holdup = Holdup0D()
    m.fs.gibbs._set_geometry()
    m.fs.gibbs._make_performance()

    assert hasattr(m.fs.gibbs, "component_list")
    assert hasattr(m.fs.gibbs, "element_list")
    assert hasattr(m.fs.gibbs, "lagrange_mult")
    assert hasattr(m.fs.gibbs, "gibbs_minimization")


def test_make_performance_has_heat_transfer_true(m):
    super(GibbsReactorData, m.fs.gibbs).build()
    m.fs.gibbs.holdup = Holdup0D()
    m.fs.gibbs._set_geometry()
    m.fs.gibbs._make_performance()

    assert hasattr(m.fs.gibbs, "heat")


def test_make_performance_has_heat_transfer_false(m):
    m.fs.gibbs.config.has_heat_transfer = False
    super(GibbsReactorData, m.fs.gibbs).build()
    m.fs.gibbs.holdup = Holdup0D()
    m.fs.gibbs._set_geometry()
    m.fs.gibbs._make_performance()

    assert not hasattr(m.fs.gibbs, "heat")


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.gibbs = GibbsReactor(property_package=m2.fs.props)
    return m2


def test_post_transform_build_SISO(m2):
    m2.fs.gibbs.post_transform_build()

    assert hasattr(m2.fs.gibbs, "inlet")
    assert hasattr(m2.fs.gibbs, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.gibbs.inlet:
        assert k in [0]
    for k in m2.fs.gibbs.outlet:
        assert k in [0]


def test_post_transform_build_inlet_list(m2):
    m2.fs.gibbs.config.inlet_list = ["a", "b"]
    m2.fs.gibbs.post_transform_build()

    assert hasattr(m2.fs.gibbs, "inlet")
    assert hasattr(m2.fs.gibbs, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.gibbs.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.gibbs.outlet:
        assert k in [0]


def test_post_transform_build_num_inlets(m2):
    m2.fs.gibbs.config.num_inlets = 2
    m2.fs.gibbs.post_transform_build()

    assert hasattr(m2.fs.gibbs, "inlet")
    assert hasattr(m2.fs.gibbs, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.gibbs.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.gibbs.outlet:
        assert k in [0]


def test_post_transform_build_outlet_list(m2):
    m2.fs.gibbs.config.outlet_list = ["a", "b"]
    m2.fs.gibbs.post_transform_build()

    assert hasattr(m2.fs.gibbs, "inlet")
    assert hasattr(m2.fs.gibbs, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.gibbs.inlet:
        assert k in [0]
    for k in m2.fs.gibbs.outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_outlet2(m2):
    m2.fs.gibbs.config.num_outlets = 2
    m2.fs.gibbs.post_transform_build()

    assert hasattr(m2.fs.gibbs, "inlet")
    assert hasattr(m2.fs.gibbs, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.gibbs.inlet:
        assert k in [0]
    for k in m2.fs.gibbs.outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.gibbs = GibbsReactor(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["H2"].fix(10.0)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["N2"].fix(47.4)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["O2"].fix(12.6)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["CO2"].fix(0.0)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["CH4"].fix(30.0)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["CO"].fix(0.0)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["H2O"].fix(0.0)
    m.fs.gibbs.inlet[:].vars["flow_mol_comp"]["NH3"].fix(0.0)
    m.fs.gibbs.inlet[:].vars["temperature"].fix(298.15)
    m.fs.gibbs.inlet[:].vars["pressure"].fix(101325.0)

    m.fs.gibbs.heat.fix(0.0)

    assert degrees_of_freedom(m) == 0

    m.fs.gibbs.initialize(state_args={
            'temperature': 298.15,
            'pressure': 101325.0,
            'flow_mol_comp': {'CH4': 11.6,
                              'CO': 16.8,
                              'CO2': 1.6,
                              'H2': 30.0,
                              'H2O': 2.9,
                              'N2': 42.8,
                              'NH3': 9.3,
                              'O2': 1.1}})

    assert (pytest.approx(11.559, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["CH4"].value)
    assert (pytest.approx(16.806, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["CO"].value)
    assert (pytest.approx(1.635, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["CO2"].value)
    assert (pytest.approx(30.047, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["H2"].value)
    assert (pytest.approx(2.924, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(42.763, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["N2"].value)
    assert (pytest.approx(9.274, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["NH3"].value)
    assert (pytest.approx(1.100, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["flow_mol_comp"]["O2"].value)
    assert (pytest.approx(298.750, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["temperature"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.gibbs.outlet[0].vars["pressure"].value)
