##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Heat Exchanger unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D
from idaes.models.heat_exchanger import HeatExchanger, HeatExchangerData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'steam_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("HeatExchangerFrame")
class _HeatExchanger(HeatExchangerData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.hx = HeatExchangerFrame(side_1_property_package=m.fs.props,
                                 side_2_property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.hx.config) == 25
    assert m.fs.hx.config.dynamic == 'use_parent_value'
    assert not m.fs.hx.config.include_holdup
    assert m.fs.hx.config.material_balance_type == 'component_phase'
    assert m.fs.hx.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.hx.config.momentum_balance_type == 'pressure'
    assert not m.fs.hx.config.has_rate_reactions
    assert not m.fs.hx.config.has_equilibrium_reactions
    assert m.fs.hx.config.has_phase_equilibrium
    assert not m.fs.hx.config.has_mass_transfer
    assert m.fs.hx.config.has_heat_transfer
    assert not m.fs.hx.config.has_work_transfer
    assert not m.fs.hx.config.has_pressure_change

    assert m.fs.hx.config.side_1_property_package == m.fs.props
    assert m.fs.hx.config.side_1_property_package_args == {}
    assert m.fs.hx.config.side_1_inlet_list is None
    assert m.fs.hx.config.side_1_num_inlets is None
    assert m.fs.hx.config.side_1_outlet_list is None
    assert m.fs.hx.config.side_1_num_outlets is None

    assert m.fs.hx.config.side_2_property_package == m.fs.props
    assert m.fs.hx.config.side_2_property_package_args == {}
    assert m.fs.hx.config.side_2_inlet_list is None
    assert m.fs.hx.config.side_2_num_inlets is None
    assert m.fs.hx.config.side_2_outlet_list is None
    assert m.fs.hx.config.side_2_num_outlets is None

    assert m.fs.hx.config.flow_type == 'counter-current'


def test_super_build(m):
    super(HeatExchangerData, m.fs.hx).build()
    assert not m.fs.hx.config.dynamic


def test_adding_holdup(m):
    super(HeatExchangerData, m.fs.hx).build()
    m.fs.hx.side_1 = Holdup0D(
            property_package=m.fs.hx.config.side_1_property_package,
            property_package_args=m.fs.hx.config.side_1_property_package_args)

    m.fs.hx.side_2 = Holdup0D(
            property_package=m.fs.hx.config.side_2_property_package,
            property_package_args=m.fs.hx.config.side_2_property_package_args)

    assert not m.fs.hx.side_1.config.include_holdup
    assert m.fs.hx.side_1.config.material_balance_type == 'component_phase'
    assert m.fs.hx.side_1.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.hx.side_1.config.momentum_balance_type == 'pressure'
    assert not m.fs.hx.side_1.config.has_rate_reactions
    assert not m.fs.hx.side_1.config.has_equilibrium_reactions
    # Properties do not have phase equilibrium , so this should be False
    assert not m.fs.hx.side_1.config.has_phase_equilibrium
    assert not m.fs.hx.side_1.config.has_mass_transfer
    assert m.fs.hx.side_1.config.has_heat_transfer
    assert not m.fs.hx.side_1.config.has_work_transfer
    assert not m.fs.hx.side_1.config.has_pressure_change
    assert m.fs.hx.side_1.config.property_package == m.fs.props
    assert m.fs.hx.side_1.config.property_package_args == {}

    assert not m.fs.hx.side_2.config.include_holdup
    assert m.fs.hx.side_2.config.material_balance_type == 'component_phase'
    assert m.fs.hx.side_2.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.hx.side_2.config.momentum_balance_type == 'pressure'
    assert not m.fs.hx.side_2.config.has_rate_reactions
    assert not m.fs.hx.side_2.config.has_equilibrium_reactions
    # Properties do not have phase equilibrium , so this should be False
    assert not m.fs.hx.side_2.config.has_phase_equilibrium
    assert not m.fs.hx.side_2.config.has_mass_transfer
    assert m.fs.hx.side_2.config.has_heat_transfer
    assert not m.fs.hx.side_2.config.has_work_transfer
    assert not m.fs.hx.side_2.config.has_pressure_change
    assert m.fs.hx.side_2.config.property_package == m.fs.props
    assert m.fs.hx.side_2.config.property_package_args == {}


def test_set_geometry_include_holdup_false(m):
    super(HeatExchangerData, m.fs.hx).build()
    m.fs.hx.side_1 = Holdup0D(
            property_package=m.fs.hx.config.side_1_property_package,
            property_package_args=m.fs.hx.config.side_1_property_package_args)
    m.fs.hx.side_2 = Holdup0D(
            property_package=m.fs.hx.config.side_2_property_package,
            property_package_args=m.fs.hx.config.side_2_property_package_args)
    m.fs.hx._set_geometry()

    assert not hasattr(m.fs.hx, "volume_side_1")
    assert not hasattr(m.fs.hx, "volume_side_2")


def test_set_geometry_include_holdup_true(m):
    m.fs.hx.config.include_holdup = True
    super(HeatExchangerData, m.fs.hx).build()
    m.fs.hx.side_1 = Holdup0D(
            property_package=m.fs.hx.config.side_1_property_package,
            property_package_args=m.fs.hx.config.side_1_property_package_args)
    m.fs.hx.side_2 = Holdup0D(
            property_package=m.fs.hx.config.side_2_property_package,
            property_package_args=m.fs.hx.config.side_2_property_package_args)
    m.fs.hx._set_geometry()

    assert hasattr(m.fs.hx, "volume_side_1")
    assert hasattr(m.fs.hx.side_1, "volume")
    assert m.fs.hx.volume_side_1 == m.fs.hx.side_1.volume

    assert hasattr(m.fs.hx, "volume_side_2")
    assert hasattr(m.fs.hx.side_2, "volume")
    assert m.fs.hx.volume_side_2 == m.fs.hx.side_2.volume


def test_make_performance(m):
    super(HeatExchangerData, m.fs.hx).build()
    m.fs.hx.side_1 = Holdup0D(
            property_package=m.fs.hx.config.side_1_property_package,
            property_package_args=m.fs.hx.config.side_1_property_package_args)

    m.fs.hx.side_2 = Holdup0D(
            property_package=m.fs.hx.config.side_2_property_package,
            property_package_args=m.fs.hx.config.side_2_property_package_args)
    m.fs.hx._set_geometry()
    m.fs.hx._make_performance()

    assert hasattr(m.fs.hx, "heat")
    assert hasattr(m.fs.hx, "heat_transfer_area")
    assert hasattr(m.fs.hx, "temperature_driving_force")
    assert hasattr(m.fs.hx, "side_1_inlet_dT")
    assert hasattr(m.fs.hx, "side_1_outlet_dT")
    assert hasattr(m.fs.hx, "heat_transfer_coefficient")
    assert hasattr(m.fs.hx, "energy_balance")
    assert hasattr(m.fs.hx, "heat_transfer_correlation")
    assert hasattr(m.fs.hx, "LMTD")


def test_make_counter_current(m):
    super(HeatExchangerData, m.fs.hx).build()
    m.fs.hx.side_1 = Holdup0D(
            property_package=m.fs.hx.config.side_1_property_package,
            property_package_args=m.fs.hx.config.side_1_property_package_args)

    m.fs.hx.side_2 = Holdup0D(
            property_package=m.fs.hx.config.side_2_property_package,
            property_package_args=m.fs.hx.config.side_2_property_package_args)
    m.fs.hx._set_geometry()
    m.fs.hx._make_performance()
    m.fs.hx._make_counter_current()

    assert hasattr(m.fs.hx, "temperature_difference_1")
    assert hasattr(m.fs.hx, "temperature_difference_2")


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.hx = HeatExchanger(side_1_property_package=m2.fs.props,
                             side_2_property_package=m2.fs.props)
    return m2


def test_post_transform_build_no_args(m2):
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_1_inlet_list(m2):
    m2.fs.hx.config.side_1_inlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_1_num_inlets(m2):
    m2.fs.hx.config.side_1_num_inlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_2_inlet_list(m2):
    m2.fs.hx.config.side_2_inlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_2_num_inlets(m2):
    m2.fs.hx.config.side_2_num_inlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_1_outlet_list(m2):
    m2.fs.hx.config.side_1_outlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_1_num_outlet2(m2):
    m2.fs.hx.config.side_1_num_outlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [0]


def test_post_transform_build_side_2_outlet_list(m2):
    m2.fs.hx.config.side_2_outlet_list = ["a", "b"]
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_side_2_num_outlet2(m2):
    m2.fs.hx.config.side_2_num_outlets = 2
    m2.fs.hx.post_transform_build()

    assert hasattr(m2.fs.hx, "side_1_inlet")
    assert hasattr(m2.fs.hx, "side_1_outlet")
    assert hasattr(m2.fs.hx, "side_2_inlet")
    assert hasattr(m2.fs.hx, "side_2_outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.hx.side_1_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_1_outlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_inlet:
        assert k in [0]
    for k in m2.fs.hx.side_2_outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.hx = HeatExchanger(side_1_property_package=m.fs.props,
                            side_2_property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.hx.side_1_inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs.hx.side_1_inlet[:].vars["temperature"].fix(866.5)
    m.fs.hx.side_1_inlet[:].vars["pressure"].fix(2.97e7)

    m.fs.hx.side_2_inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs.hx.side_2_inlet[:].vars["temperature"].fix(500.0)
    m.fs.hx.side_2_inlet[:].vars["pressure"].fix(2.97e7)

    m.fs.hx.heat_transfer_coefficient.fix(5.0)
    m.fs.hx.heat_transfer_area.fix(100000.0)

    assert degrees_of_freedom(m) == 0

    m.fs.hx.initialize()

    assert (pytest.approx(27500, abs=1e-2) ==
            m.fs.hx.side_1_outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(789.0, abs=1e-2) ==
            m.fs.hx.side_1_outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.hx.side_1_outlet[0].vars["pressure"].value)

    assert (pytest.approx(27500, abs=1e-2) ==
            m.fs.hx.side_2_outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(529.789, abs=1e-2) ==
            m.fs.hx.side_2_outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.hx.side_2_outlet[0].vars["pressure"].value)
