##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Mixer unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup0D, HoldupStatic
from idaes.models.mixer import Mixer, MixerData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'steam_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("MixerFrame")
class _Mixer(MixerData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.mix = MixerFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.mix.config) == 18
    assert not m.fs.mix.config.dynamic
    assert not m.fs.mix.config.include_holdup
    assert m.fs.mix.config.material_balance_type == 'component_phase'
    assert m.fs.mix.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.mix.config.momentum_balance_type == 'pressure'
    assert not m.fs.mix.config.has_rate_reactions
    assert not m.fs.mix.config.has_equilibrium_reactions
    assert not m.fs.mix.config.has_phase_equilibrium
    assert not m.fs.mix.config.has_mass_transfer
    assert not m.fs.mix.config.has_heat_transfer
    assert not m.fs.mix.config.has_work_transfer
    assert not m.fs.mix.config.has_pressure_change
    assert m.fs.mix.config.property_package == m.fs.props
    assert m.fs.mix.config.property_package_args == {}
    assert m.fs.mix.config.inlet_list is None
    assert m.fs.mix.config.num_inlets is None
    assert m.fs.mix.config.outlet_list is None
    assert m.fs.mix.config.num_outlets is None


def test_super_build(m):
    super(MixerData, m.fs.mix).build()
    assert not m.fs.mix.config.dynamic


# -----------------------------------------------------------------------------
# Testing build method
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.mix = Mixer(property_package=m2.fs.props)
    return m2


def test_build_validate_inlet_args(m2):
    assert m2.fs.mix.config.num_inlets == 2


def test_build_inlet_list():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.mix = Mixer(property_package=m.fs.props,
                     inlet_list=["1", "2"])

    assert m.fs.mix.config.inlet_list == ["1", "2"]
    assert m.fs.mix.config.num_inlets is None


def test_build_num_inlets():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.mix = Mixer(property_package=m.fs.props,
                     num_inlets=3)

    assert m.fs.mix.config.inlet_list is None
    assert m.fs.mix.config.num_inlets == 3


def test_build_check_HoldupStatic(m2):
    assert isinstance(m2.fs.mix.holdup, HoldupStatic)


def test_build_check_Holdup0D():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.mix = Mixer(property_package=m.fs.props,
                     include_holdup=True)

    assert isinstance(m.fs.mix.holdup, Holdup0D)


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
def test_post_transform_build_no_args(m2):
    m2.fs.mix.post_transform_build()

    assert hasattr(m2.fs.mix, "inlet")
    assert hasattr(m2.fs.mix, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.mix.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.mix.outlet:
        assert k in [0]


def test_post_transform_build_inlet_list(m2):
    m2.fs.mix.config.inlet_list = ["a", "b"]
    m2.fs.mix.post_transform_build()

    assert hasattr(m2.fs.mix, "inlet")
    assert hasattr(m2.fs.mix, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.mix.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.mix.outlet:
        assert k in [0]


def test_post_transform_build_num_inlets(m2):
    m2.fs.mix.config.num_inlets = 2
    m2.fs.mix.post_transform_build()

    assert hasattr(m2.fs.mix, "inlet")
    assert hasattr(m2.fs.mix, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.mix.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.mix.outlet:
        assert k in [0]


def test_post_transform_build_outlet_list(m2):
    m2.fs.mix.config.outlet_list = ["a", "b"]
    m2.fs.mix.post_transform_build()

    assert hasattr(m2.fs.mix, "inlet")
    assert hasattr(m2.fs.mix, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.mix.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.mix.outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_outlet2(m2):
    m2.fs.mix.config.num_outlets = 2
    m2.fs.mix.post_transform_build()

    assert hasattr(m2.fs.mix, "inlet")
    assert hasattr(m2.fs.mix, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.mix.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.mix.outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.mix = Mixer(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.mix.inlet[:, "1"].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs.mix.inlet[:, "1"].vars["temperature"].fix(866.5)
    m.fs.mix.inlet[:, "1"].vars["pressure"].fix(2.97e7)
    m.fs.mix.inlet[:, "2"].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs.mix.inlet[:, "2"].vars["temperature"].fix(866.5)
    m.fs.mix.inlet[:, "2"].vars["pressure"].fix(2.97e7)

    assert degrees_of_freedom(m) == 0

    m.fs.mix.initialize()

    assert (pytest.approx(55000.0, abs=1e-2) ==
            m.fs.mix.outlet[0].vars["flow_mol_comp"]["H2O"].value)
    assert (pytest.approx(866.5, abs=1e-2) ==
            m.fs.mix.outlet[0].vars["temperature"].value)
    assert (pytest.approx(29700000.0, abs=1e-2) ==
            m.fs.mix.outlet[0].vars["pressure"].value)
