##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for PFR unit model.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, SolverFactory
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        Holdup1D
from idaes.models.plug_flow_reactor import PFR, PFRData
from idaes.ui.report import degrees_of_freedom

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
               'core', 'examples', 'reaction_properties_demo.py')
fe = import_file(example)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("PFRFrame")
class _PFR(PFRData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.pfr = PFRFrame(property_package=m.fs.props)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert len(m.fs.pfr.config) == 24
    assert m.fs.pfr.config.dynamic == 'use_parent_value'
    assert not m.fs.pfr.config.include_holdup
    assert m.fs.pfr.config.material_balance_type == 'component_phase'
    assert m.fs.pfr.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.pfr.config.momentum_balance_type == 'pressure'
    assert m.fs.pfr.config.has_rate_reactions
    assert m.fs.pfr.config.has_equilibrium_reactions
    assert not m.fs.pfr.config.has_phase_equilibrium
    assert not m.fs.pfr.config.has_mass_transfer
    assert not m.fs.pfr.config.has_heat_transfer
    assert not m.fs.pfr.config.has_work_transfer
    assert not m.fs.pfr.config.has_pressure_change
    assert m.fs.pfr.config.property_package == m.fs.props
    assert m.fs.pfr.config.property_package_args == {}
    assert m.fs.pfr.config.inlet_list is None
    assert m.fs.pfr.config.num_inlets is None
    assert m.fs.pfr.config.outlet_list is None
    assert m.fs.pfr.config.num_outlets is None

    assert m.fs.pfr.config.discretization_method == "OCLR"
    assert m.fs.pfr.config.finite_elements == 20
    assert m.fs.pfr.config.collocation_points == 3

    assert not m.fs.pfr.config.has_mass_diffusion
    m.fs.pfr.config.has_mass_diffusion = True
    m.fs.pfr.config.has_mass_diffusion = False
    with pytest.raises(ValueError):
        m.fs.pfr.config.has_mass_diffusion = 'foo'
    with pytest.raises(ValueError):
        m.fs.pfr.config.has_mass_diffusion = 10

    assert not m.fs.pfr.config.has_energy_diffusion
    m.fs.pfr.config.has_energy_diffusion = True
    m.fs.pfr.config.has_energy_diffusion = False
    with pytest.raises(ValueError):
        m.fs.pfr.config.has_energy_diffusion = 'foo'
    with pytest.raises(ValueError):
        m.fs.pfr.config.has_energy_diffusion = 10

    assert m.fs.pfr.config.velocity_type == 'none'


def test_super_build(m):
    super(PFRData, m.fs.pfr).build()
    assert not m.fs.pfr.config.dynamic


def test_adding_holdup(m):
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()

    assert not m.fs.pfr.holdup.config.include_holdup
    assert m.fs.pfr.holdup.config.material_balance_type == 'component_phase'
    assert m.fs.pfr.holdup.config.energy_balance_type == 'enthalpy_total'
    assert m.fs.pfr.holdup.config.momentum_balance_type == 'pressure'
    assert m.fs.pfr.holdup.config.has_rate_reactions
    # Properties do not have equilibrium reactions, so this should be False
    assert not m.fs.pfr.holdup.config.has_equilibrium_reactions
    assert not m.fs.pfr.holdup.config.has_phase_equilibrium
    assert not m.fs.pfr.holdup.config.has_mass_transfer
    assert not m.fs.pfr.holdup.config.has_heat_transfer
    assert not m.fs.pfr.holdup.config.has_work_transfer
    assert not m.fs.pfr.holdup.config.has_pressure_change
    assert m.fs.pfr.holdup.config.property_package == m.fs.props
    assert m.fs.pfr.holdup.config.property_package_args == {}
    assert m.fs.pfr.holdup.config.discretization_method == "OCLR"
    assert m.fs.pfr.holdup.config.finite_elements == 20
    assert m.fs.pfr.holdup.config.collocation_points == 3
    assert not m.fs.pfr.holdup.config.has_mass_diffusion
    assert not m.fs.pfr.holdup.config.has_energy_diffusion


def test_set_geometry_has_velocity_false(m):
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()

    assert hasattr(m.fs.pfr, "volume")
    assert hasattr(m.fs.pfr, "area")
    assert hasattr(m.fs.pfr, "length")
    assert hasattr(m.fs.pfr, "ldomain")
    assert not hasattr(m.fs.pfr, "velocity")
    assert not hasattr(m.fs.pfr, "velocity_calculation")
    assert not hasattr(m.fs.pfr, "link_velocity")


def test_set_geometry_velocity_type_total(m):
    m.fs.pfr.config.velocity_type = 'mixture'
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()

    assert hasattr(m.fs.pfr, "volume")
    assert hasattr(m.fs.pfr, "area")
    assert hasattr(m.fs.pfr, "length")
    assert hasattr(m.fs.pfr, "ldomain")
    assert hasattr(m.fs.pfr, "velocity")
    assert hasattr(m.fs.pfr.holdup, "velocity_calculation")
    assert not hasattr(m.fs.pfr, "link_velocity")


def test_set_geometry_velocity_type_total_link(m):
    m.fs.pfr.config.velocity_type = 'mixture'
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr.holdup.properties[0, 1].velocity = 1
    m.fs.pfr._set_geometry()

    assert hasattr(m.fs.pfr, "volume")
    assert hasattr(m.fs.pfr, "area")
    assert hasattr(m.fs.pfr, "length")
    assert hasattr(m.fs.pfr, "ldomain")
    assert hasattr(m.fs.pfr, "velocity")
    assert hasattr(m.fs.pfr.holdup, "velocity_calculation")
    assert hasattr(m.fs.pfr, "link_velocity")


def test_make_performance_rate_reactions_true(m):
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()
    m.fs.pfr._make_performance()

    assert hasattr(m.fs.pfr, "rate_reaction_idx")
    assert hasattr(m.fs.pfr, "rate_reaction_extents")


def test_make_performance_rate_reactions_false(m):
    m.fs.pfr.config.has_rate_reactions = False
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()
    m.fs.pfr._make_performance()

    assert not hasattr(m.fs.pfr, "rate_reaction_idx")
    assert not hasattr(m.fs.pfr, "rate_reaction_extents")


def test_make_performance_has_heat_transfer_true(m):
    m.fs.pfr.config.has_heat_transfer = True
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()
    m.fs.pfr._make_performance()

    assert hasattr(m.fs.pfr, "heat")


def test_make_performance_has_heat_transfer_false(m):
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()
    m.fs.pfr._make_performance()

    assert not hasattr(m.fs.pfr, "heat")


def test_make_performance_has_pressure_change_true(m):
    m.fs.pfr.config.has_pressure_change = True
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()
    m.fs.pfr._make_performance()

    assert hasattr(m.fs.pfr, "deltaP")


def test_make_performance_has_pressure_change_false(m):
    super(PFRData, m.fs.pfr).build()
    m.fs.pfr.holdup = Holdup1D()
    m.fs.pfr._set_geometry()
    m.fs.pfr._make_performance()

    assert not hasattr(m.fs.pfr, "deltaP")


# -----------------------------------------------------------------------------
# Test transformation of length domain
def test_OCLR():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.pfr = PFR(property_package=m.fs.props)

    assert len(m.fs.pfr.ldomain) == 61


def test_OCLL():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.pfr = PFR(property_package=m.fs.props,
                   discretization_method='OCLL',
                   finite_elements=5,
                   collocation_points=4)

    assert len(m.fs.pfr.ldomain) == 21


def test_BFD():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.pfr = PFR(property_package=m.fs.props,
                   discretization_method='BFD',
                   finite_elements=6)

    assert len(m.fs.pfr.ldomain) == 7


def test_FFD():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()
    m.fs.pfr = PFR(property_package=m.fs.props,
                   discretization_method='FFD',
                   finite_elements=9)

    assert len(m.fs.pfr.ldomain) == 10


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
@pytest.fixture()
def m2():
    """
    Build a simple flowsheet model for build tests from.
    """
    m2 = ConcreteModel()
    m2.fs = Flowsheet(dynamic=False)
    m2.fs.props = fe.PropertyParameterBlock()
    m2.fs.pfr = PFR(property_package=m2.fs.props)
    return m2


def test_post_transform_build_SISO(m2):
    m2.fs.pfr.post_transform_build()

    assert hasattr(m2.fs.pfr, "inlet")
    assert hasattr(m2.fs.pfr, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.pfr.inlet:
        assert k in [0]
    for k in m2.fs.pfr.outlet:
        assert k in [0]


def test_post_transform_build_inlet_list(m2):
    m2.fs.pfr.config.inlet_list = ["a", "b"]
    m2.fs.pfr.post_transform_build()

    assert hasattr(m2.fs.pfr, "inlet")
    assert hasattr(m2.fs.pfr, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.pfr.inlet:
        assert k in [(0, "a"), (0, "b")]
    for k in m2.fs.pfr.outlet:
        assert k in [0]


def test_post_transform_build_num_inlets(m2):
    m2.fs.pfr.config.num_inlets = 2
    m2.fs.pfr.post_transform_build()

    assert hasattr(m2.fs.pfr, "inlet")
    assert hasattr(m2.fs.pfr, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.pfr.inlet:
        assert k in [(0, "1"), (0, "2")]
    for k in m2.fs.pfr.outlet:
        assert k in [0]


def test_post_transform_build_outlet_list(m2):
    m2.fs.pfr.config.outlet_list = ["a", "b"]
    m2.fs.pfr.post_transform_build()

    assert hasattr(m2.fs.pfr, "inlet")
    assert hasattr(m2.fs.pfr, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.pfr.inlet:
        assert k in [0]
    for k in m2.fs.pfr.outlet:
        assert k in [(0, "a"), (0, "b")]


def test_post_transform_build_num_outlet2(m2):
    m2.fs.pfr.config.num_outlets = 2
    m2.fs.pfr.post_transform_build()

    assert hasattr(m2.fs.pfr, "inlet")
    assert hasattr(m2.fs.pfr, "outlet")

    # Test that inlet and outlet have correct indexing sets
    for k in m2.fs.pfr.inlet:
        assert k in [0]
    for k in m2.fs.pfr.outlet:
        assert k in [(0, "1"), (0, "2")]


# -----------------------------------------------------------------------------
# Test initialization
# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None


@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialization():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props = fe.PropertyParameterBlock()

    m.fs.pfr = PFR(property_package=m.fs.props)
    m.fs.post_transform_build()

    m.fs.pfr.inlet[:].vars["flow_mol_comp"]["a"].fix(1.0)
    m.fs.pfr.inlet[:].vars["flow_mol_comp"]["b"].fix(2.0)
    m.fs.pfr.inlet[:].vars["flow_mol_comp"]["c"].fix(0.1)
    m.fs.pfr.inlet[:].vars["flow_mol_comp"]["d"].fix(0.0)
    m.fs.pfr.inlet[:].vars["flow_mol_comp"]["e"].fix(0.0)
    m.fs.pfr.inlet[:].vars["flow_mol_comp"]["f"].fix(0.0)
    m.fs.pfr.inlet[:].vars["enth_mol"].fix(-238.71)
    m.fs.pfr.inlet[:].vars["pressure"].fix(101325.0)

    m.fs.pfr.area.fix(1.0)
    m.fs.pfr.length.fix(10.0)

    assert degrees_of_freedom(m) == 0

    m.fs.pfr.initialize()

    assert (pytest.approx(0.609, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["flow_mol_comp"]["a"].value)
    assert (pytest.approx(1.464, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["flow_mol_comp"]["b"].value)
    assert (pytest.approx(0.244, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["flow_mol_comp"]["c"].value)
    assert (pytest.approx(0.146, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["flow_mol_comp"]["d"].value)
    assert (pytest.approx(0.002, abs=1e-3) ==
            m.fs.pfr.outlet[0].vars["flow_mol_comp"]["e"].value)
    assert (pytest.approx(0.244, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["flow_mol_comp"]["f"].value)
    assert (pytest.approx(-273.147, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["enth_mol"].value)
    assert (pytest.approx(101325.0, abs=1e-2) ==
            m.fs.pfr.outlet[0].vars["pressure"].value)
