##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
from __future__ import absolute_import, division, print_function

from pyomo.core.base.symbolic import *
from pyomo.core.base.symbolic import _functionMap

from pyomo.environ import ExternalFunction
from pyomo.core.base.constraint import _ConstraintData, Constraint
from pyomo.core.base.expression import _ExpressionData
from pyomo.core.base.block import _BlockData
from pyomo.core.expr.expr_pyomo5 import (
    SimpleExpressionVisitor,
    ExternalFunctionExpression,
    ExpressionBase)
import logging

#TODO<jce> This is based on Pyomo's symbolic module, which is being updated
#          see PR https://github.com/Pyomo/pyomo/pull/719, (more efficent)

#TODO<jce> Look into things like sum operator and template expressions

class SympifyVisitorWithFunc(SympifyVisitor):
    """
    This is based on the class in Pyomo's symbolic module, but also replaces
    ExternalFunctions.
    """
    def __init__(self, *args, **kwargs):
        super(SympifyVisitorWithFunc, self).__init__(*args, **kwargs)
        self.seen = set()

    def visit(self, node, values):
        if node.__class__ is EXPR.UnaryFunctionExpression:
            self.seen.add(id(node))
            return _functionMap[node._name](values[0])
        elif node.__class__ is EXPR.ExternalFunctionExpression:
            self.seen.add(id(node._fcn))
            return self.pyomo2sympy[id(node._fcn)](*values)
        else:
            self.seen.add(id(node))
            return node._apply_operation(values)

    def visiting_potential_leaf(self, node):
        self.seen.add(id(node))
        # Don't replace native or sympy types
        if type(node) in self.native_or_sympy_types:
            return True, node
        # Replace pyomo variables with sympy variables
        if id(node) in self.pyomo2sympy:
            return True, self.pyomo2sympy[id(node)]
        # Replace constants
        if not node.is_potentially_variable():
            return True, value(node)
        # Don't replace anything else
        return False, None

    def finalize(self, ans):
        return (ans, self.seen)


class _NamedExpressionVisitor(EXPR.SimpleExpressionVisitor):
    """
    Visitor to find named expressions in a Pyomo expression
    """
    def __init__(self):
        super(_NamedExpressionVisitor, self).__init__()
        self.seen = set()  # components that have already been added
        self.comps = []    # components that are named expressions

    def visit(self, node):
        try:
            if node.is_named_expression_type():
                if id(node) in self.seen:
                    return
                self.seen.add(id(node))
                self.comps.append(node)
                return node
        except AttributeError:
            pass #doesn't have is_named_expression_type, so not named expr

    def finalize(self):
        return self.comps


class _ExternalFunctionVisitor(EXPR.SimpleExpressionVisitor):
    """
    Visitor to find external functions
    """
    def __init__(self):
        super(_ExternalFunctionVisitor, self).__init__()
        self.seen = set() # Functions that have already been added
        self.comps = []   # List of external functions in expression

    def visit(self, node):
        if isinstance(node, ExternalFunctionExpression):
            if id(node) in self.seen:
                return
            self.seen.add(id(node))
            self.comps.append(node._fcn)
            return node

    def finalize(self):
        return self.comps


def identify_named_expressions(expr):
    """
    Args:
        expr: a Pyomo expression in which to look for named expressions

    Returns:
        List of named expressions in expr
    """
    visitor = _NamedExpressionVisitor()
    ne = visitor.xbfs(expr)
    # If the root node is a named expression, I'll assume you do want to
    # descend into it, so I don't want to it in the list of things to replace
    if expr in ne: ne.remove(expr)
    return ne

def identify_external_functions(expr):
    """
    Args:
        expr: a Pyomo expression in which to look for ExternalFunctions

    Returns:
        List of named expressions in expr
    """
    visitor = _ExternalFunctionVisitor()
    return visitor.xbfs(expr)

def deduplicate_symbol(x, v, used):
    """
    Check if x is a duplicated LaTeX symbol if so add incrimenting <i>

    Args:
        x: symbol string
        v: pyomo object
        used: dictionary of pyomo objects with symbols as keys

    Returns:
        Returns a unique symbol.  If x was not in used keys, returns x,
        otherwise adds exponents to make it unique.
    """
    y = x
    k = 1
    while x in used and id(used[x]) != id(v):
        x = "{} \\langle {} \\rangle".format(y, k)
        k += 1
    used[x] = v
    return x

def sympify_expression_with_func(expr, sympySymbols, pyomo2sympy):
    """
    This function converts a Pyomo expression to a sympy expression, it is based
    of the one in Pyomo's symbolic module but also replaces named expressions.
    Not expanding named expressions leads to much more readable results.
    """
    native_or_sympy_types = set(native_types)
    native_or_sympy_types.add( type(list(sympySymbols)[0]) )
    visitor = SympifyVisitorWithFunc(native_or_sympy_types, pyomo2sympy)
    return visitor.dfs_postorder_stack(expr)

def _add_docs(pyomo_comps, sympy_comps, docs, seen, head, symbol_names,
    parent_symbol):
    """
    Adds documentation for a set of pyomo components to a markdown table

    Args:
        pyomo_comps: iterable containing pyomo_components
        docs: string containing a mardown table
        seen: a set of compoents seen in the expression being documented

    Returns:
        A new string markdown table with added doc rows.
    """
    docked = set() # components already documented, mainly for idexed compoents
    nohead = 1  # write hedding before adding first item. If no items no heading
    for i, c in enumerate(pyomo_comps):
        if id(c) in seen:  # only document components actually in expression
            if nohead: # add heading if is first entry in this section
                docs += "**{}** | **Doc** | **Path**\n".format(head)
                nohead = 0
            c = c.parent_component() # Document the parent for indexed comps
            if id(c) not in docked:
                docked.add(id(c)) # make sure we don't get a line for every index
                try:
                    s = parent_symbol[id(c)][1]
                except KeyError:
                    s = symbol_names[sympy_comps[i]]
                docs += "${}$|{}|{}\n"\
                        .format(s, c.doc, c)
    return docs

def to_markdown(expr):
    """Return a sympy expression for the given Pyomo expression

    Args:
        expr (Expression): Pyomo expression
    Returns:
        sympy expression
    """
    pyomo_vars = list(EXPR.identify_variables(expr))      # list of vars
    sympy_vars = [None]*len(pyomo_vars)
    pyomo_extf = identify_external_functions(expr)
    sympy_func = [None]*len(pyomo_extf)
    pyomo_expr = identify_named_expressions(expr)
    sympy_expv = [None]*len(pyomo_expr)

    used = {}
    symbol_names = {} # Sympy symbol to LaTeX symbol
    parent_symbol = {} # Base sympy and LaTeX symbol for parent of element

    def symbol_tables(base_symbol, pyomo_objs, sympy_objs, sympy_make):
        j = 1
        for i, v in enumerate(pyomo_objs):
            o = v.parent_component()
            if o.is_indexed() and id(o) in parent_symbol:
                x = parent_symbol[id(o)][0]
                latex_symbol = parent_symbol[id(o)][1]
            else:
                x = "{}_{}".format(base_symbol, j)
                j += 1
                try:
                    latex_symbol = deduplicate_symbol(o.latex_symbol, o, used)
                except:
                    latex_symbol = x
            if o.is_indexed():
                parent_symbol[id(o)] = (x, latex_symbol)
                idx = v.index()
                idxl = idx
                if isinstance(idx, tuple):
                    idxl = ",".join(idx)
                    idx = "|".join(idx)
                x = "{}[{}]".format(x, idx)
                latex_symbol = "{}[{}]".format(latex_symbol, idxl)
            sympy_objs[i] = sympy_make(x)
            symbol_names[sympy_objs[i]] = latex_symbol

    symbol_tables("x", pyomo_vars, sympy_vars, sympy.var)
    symbol_tables("u", pyomo_expr, sympy_expv, sympy.var)
    symbol_tables("func", pyomo_extf, sympy_func, sympy.Function)

    pyomo_comp = pyomo_vars + pyomo_extf + pyomo_expr
    sympy_comp = sympy_vars + sympy_func + sympy_expv

    sympy2pyomo = dict( zip(sympy_comp, pyomo_comp) )
    pyomo2sympy = dict( (id(pyomo_comp[i]), sympy_comp[i])
                         for i in range(len(pyomo_comp)) )

    # A sympy expression and a list of Pyomo components seen by the vistor
    # Seen is for documentation, so can display only docs for seen components
    sympy_expr, seen = sympify_expression_with_func(expr, sympy2pyomo, pyomo2sympy)
    # This next bit documents the expression, could use a lot of work, but
    # for now it generates markdown tables that are resonably readable in a
    # jupyter notebook.
    docs = "\nSymbol | Doc | Path\n ---: | --- | ---\n"
    docs = _add_docs(pyomo_vars, sympy_vars, docs, seen, "Variable",
        symbol_names, parent_symbol)
    docs = _add_docs(pyomo_expr, sympy_expv, docs, seen, "Expression",
        symbol_names, parent_symbol)
    docs = _add_docs(pyomo_extf, sympy_func, docs, seen, "Function",
        symbol_names, parent_symbol)

    #probably should break this up, but this will do for now.
    return {"sympy_expr": sympy_expr,
            "where":docs,
            "latex_expr":sympy.latex(sympy_expr, symbol_names=symbol_names)}

def document_constraints(comp, doc=True, descend_into=True):
    """
    Provides nicely formatted constraint documetntation in markdown format,
    assuming the $$latex math$$ and $latex math$ syntax is supported.

    Args:
        comp: A Pyomo component to document in {_ConstraintData, _ExpressionData,
                _BlockData}.
        doc: True adds a documentation table for each constraint or expression.
                Due to the way symbols are semi-automatiaclly generated, the
                exact symbol definitions may be unique to each constraint or
                expression, if unique LaTeX symbols were not provided
                everywhere in a block.
        descend_into: If True, look in subblocks for constraints.

    Returns:
        A string in markdown format with equations in LaTeX form.
    """
    if isinstance(comp, _ExpressionData):
        d = to_markdown(comp)
        if doc:
            return "$${}$$\n{}".format(d["latex_expr"], d["where"])
        else:
            return "$${}$$".format(d["latex_expr"])
    elif isinstance(comp, _ConstraintData):
        d = to_markdown(comp.body)
        if doc:
            return "$${} \le {}\le {}$$\n{}".format(
                comp.lower, d["latex_expr"], comp.upper, d["where"])
        else:
            return "$${} \le {}\le {}$$".format(
                comp.lower, d["latex_expr"], comp.upper)
    elif isinstance(comp, _BlockData):
        cs = []
        for c in comp.component_data_objects(
            Constraint, descend_into=descend_into):
            if not c.active:
                continue
            cs.append("## {}".format(c))
            cs.append(document_constraints(c, doc))
        return "\n".join(cs)
