#!/usr/bin/env bash
#
# Unix installation script for IDAES toolkit
#

function idaes_usage () {
  printf "\n"
  printf "usage: $0 [-hn] name\n"
  printf "  name  Name of the Conda environment to use.\n"
  printf "        This environment will be created if it does\n"
  printf "        not already exist.\n"
  printf "   -h   Show this help message\n"
  printf "   -n   Dry run, do nothing (always succeeds)\n"
}

function idaes_install_ok () {
  printf "\n** SUCCESS **\n"
  printf "\nNow activate the conda environment to run test, build docs, etc.\n"
  printf "with the following command:\n\n"
  printf "  source activate $condaenv\n\n"
  exit 0
}

function idaes_install_fail () {
  printf "\n!! FAILURE !!\n"
  exit -1
}

function idaes_check_command () {
  name=$1
  status=$2
  if [ $status -ne 0 ]; then
    printf "ERROR: Command '%s' failed\n" "$name"
    idaes_install_fail
  fi
}

##
# Main
##

log=install_$$.log
dryrun=0
condaenv=""
while [ $# -gt 0 ]; do
   cmd=$1 ; shift
   case "$cmd" in
       -h ) idaes_usage; exit 0 ;;
       -n ) dryrun=1 ;;
       *  ) condaenv="$cmd" ;;
   esac
done

# Check args
if [ "$condaenv" = "" ]; then
  printf "ERROR: You must provide a name for the Conda environment\n"
  idaes_usage
  idaes_install_fail
fi

printf "$(date): Starting install ${cmd}\n" > ${log}

# 1. Create the conda environment
if [ $dryrun -eq 1 ]; then
  printf "... Skipping create conda environment\n" | tee -a ${log}
else
    printf "Checking for '${condaenv}' env\n" >> ${log}
    conda list --name "$condaenv" >> ${log} 2>&1
  if [ $? -eq 0 ]; then
    printf "... Using existing conda environment for '%s'\n" "$condaenv" | tee -a ${log}
  else
    printf "... Creating conda environment '%s'\n" "$condaenv" | tee -a ${log}
    conda create --yes --name "$condaenv" jupyter >> ${log}
    idaes_check_command "conda create" $?
  fi
fi

# 2. Build and install the IDAES code into the environment
if [ $dryrun -eq 1 ]; then
  printf "... Skipping build and install\n"
else
  printf "... Activate environment\n" | tee -a ${log}
  source activate "$condaenv"
  if [ "${CONDA_DEFAULT_ENV}" != "$condaenv" ]; then
    printf "ERROR: Environment is '%s' not '%s'\n" "${CONDA_DEFAULT_ENV}" "$condaenv"
    idaes_install_fail
  fi
  printf "... Build and install\n" | tee -a ${log}
  pip install -r requirements.txt >> ${log} 2>&1
  idaes_check_command "pip install requirements" $?
  python setup.py install >> ${log} 2>&1
  idaes_check_command "python install" $?
fi

# Done!
idaes_install_ok
